#include <omp.h>
#include <cmath>
#include "conversions.h"

void VEpsDel2Cij(float *c11,float *c13,float *c33,const float *v,const float *eps,const float *del,float v0,float eps0,float del0,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  c33[i]=v0*v0*v[i]*v[i]; 
  c11[i]=c33[i]*(1.+2.*eps0*eps[i]); 
  c13[i]=c33[i]*sqrt(1.+2.*del0*del[i]); 
 }
 return;
}

void GradCij2GradVEpsDel(float *gv,float *geps,float *gdel,const float *gc11,const float *gc13,const float *gc33,const float *v,const float *eps,const float *del,float v0,float eps0,float del0,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  float tv=v0*v[i];
  float teps=1.+2.*eps0*eps[i];
  float tdel=1.+2.*del0*del[i];
  gv[i]=2.*tv*(gc11[i]*teps+gc13[i]*sqrt(tdel)+gc33[i])*v0;
  geps[i]=2.*tv*tv*gc11[i]*eps0;
  gdel[i]=tv*tv/sqrt(tdel)*gc13[i]*del0;
 }
 return;
}

void VnEtaDel2Cij(float *c11,float *c13,float *c33,const float *vn,const float *eta,const float *del,float vn0,float eta0,float del0,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  float vn2=vn0*vn0*vn[i]*vn[i];
  c11[i]=vn2*(1.+2.*eta0*eta[i]); 
  c13[i]=vn2/sqrt(1.+2.*del0*del[i]); 
  c33[i]=vn2/(1.+2.*del0*del[i]); 
 }
 return;
}

void GradCij2GradVnEtaDel(float *gvn,float *geta,float *gdel,const float *gc11,const float *gc13,const float *gc33,const float *vn,const float *eta,const float *del,float vn0,float eta0,float del0,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  float tvn=vn0*vn[i];
  float teta=1.+2.*eta0*eta[i];
  float tdel=1.+2.*del0*del[i];
  gvn[i]=2.*tvn*(gc11[i]*teta+gc13[i]/sqrt(tdel)+gc33[i]/tdel)*vn0;
  geta[i]=2.*tvn*tvn*gc11[i]*eta0;
  gdel[i]=-tvn*tvn/tdel*(gc13[i]/sqrt(tdel)+2.*gc33[i]/tdel)*del0;
 }
 return;
}

void VhEpsEta2Cij(float *c11,float *c13,float *c33,const float *vh,const float *eps,const float *eta,float vh0,float eps0,float eta0,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  c11[i]=vh0*vh0*vh[i]*vh[i]; 
  c13[i]=c11[i]/sqrt((1.+2.*eps0*eps[i])*(1.+2.*eta0*eta[i])); 
  c33[i]=c11[i]/(1.+2.*eps0*eps[i]); 
 }
 return;
}

void GradCij2GradVhEpsEta(float *gvh,float *geps,float *geta,const float *gc11,const float *gc13,const float *gc33,const float *vh,const float *eps,const float *eta,float vh0,float eps0,float eta0,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  float tvh=vh0*vh[i];
  float teps=1.+2.*eps0*eps[i];
  float teta=1.+2.*eta0*eta[i];
  gvh[i]=2.*tvh*(gc11[i]+gc13[i]/sqrt(teps*teta)+gc33[i]/teps)*vh0;
  geps[i]=-tvh*tvh/teps*(gc13[i]/sqrt(teps*teta)+2.*gc33[i]/teps)*eps0;
  geta[i]=-tvh*tvh/sqrt(teps*teta)/teta*gc13[i]*eta0;
 }
 return;
}

void VVhDel2Cij(float *c11,float *c13,float *c33,const float *v,const float *vh,const float *del,float v0,float vh0,float del0,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  c11[i]=vh0*vh0*vh[i]*vh[i]; 
  c33[i]=v0*v0*v[i]*v[i]; 
  c13[i]=c33[i]*sqrt(1.+2.*del0*del[i]); 
 }
 return;
}

void GradCij2GradVVhDel(float *gv,float *gvh,float *gdel,const float *gc11,const float *gc13,const float *gc33,const float *v,const float *vh,const float *del,float v0,float vh0,float del0,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  float tv=v0*v[i];
  float tvh=vh0*vh[i];
  float tdel=1.+2.*del0*del[i];
  gv[i]=2.*tv*(gc13[i]*sqrt(tdel)+gc33[i])*v0;
  gvh[i]=2.*tvh*gc11[i]*vh0;
  gdel[i]=tv*tv/sqrt(tdel)*gc13[i]*del0;
 }
 return;
}

void VnVhDel2Cij(float *c11,float *c13,float *c33,const float *vn,const float *vh,const float *del,float vn0,float vh0,float del0,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  c11[i]=vh0*vh0*vh[i]*vh[i]; 
  c13[i]=vn0*vn[i]*vn0*vn[i]/sqrt(1.+2.*del0*del[i]); 
  c33[i]=vn0*vn[i]*vn0*vn[i]/(1.+2.*del0*del[i]); 
 }
 return;
}

void GradCij2GradVnVhDel(float *gvn,float *gvh,float *gdel,const float *gc11,const float *gc13,const float *gc33,const float *vn,const float *vh,const float *del,float vn0,float vh0,float del0,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  float tvn=vn0*vn[i];
  float tvh=vh0*vh[i];
  float tdel=1.+2.*del0*del[i];
  gvn[i]=2.*tvn*(gc13[i]/sqrt(tdel)+gc33[i]/tdel)*vn0;
  gvh[i]=2.*tvh*gc11[i]*vh0;
  gdel[i]=-tvn*tvn/tdel*(gc13[i]/sqrt(tdel)+2.*gc33[i]/tdel)*del0;
 }
 return;
}

void VVnVh2Cij(float *c11,float *c13,float *c33,const float *v,const float *vn,const float *vh,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  c11[i]=vh[i]*vh[i]*1e6; 
  c13[i]=v[i]*vn[i]*1e6; 
  c33[i]=v[i]*v[i]*1e6; 
 }
 return;
}

void GradCij2GradVVnVh(float *gv,float *gvn,float *gvh,const float *gc11,const float *gc13,const float *gc33,const float *v,const float *vn,const float *vh,int n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  gv[i]=(gc13[i]*vn[i]+gc33[i]*2.*v[i])*1e6;
  gvn[i]=(gc13[i]*v[i])*1e6;
  gvh[i]=(gc11[i]*2.*vh[i])*1e6;
 }
 return;
}
