#include <omp.h>
#include <cmath>
#include "myio.h"
#include "conversions.h"

using namespace std;

int main(int argc,char **argv){
 myio_init(argc,argv);
 int nx,nz;
 float ox,oz,dx,dz;
 get_param("nx",nx,"dx",dx,"ox",ox);
 get_param("nz",nz,"dz",dz,"oz",oz);
 int n=nx*nz;

 int direction; get_param("direction",direction);

 float *c11=new float[n]();  
 float *c13=new float[n]();  
 float *c33=new float[n]();  
 
 if(direction==1){
  read("c11",c11,n); 
  read("c13",c13,n); 
  read("c33",c33,n); 
  float *v=new float[n]();
  float *eps=new float[n]();
  float *del=new float[n]();
  #pragma omp parallel for default(shared)
  for(int i=0;i<n;++i){
   eps[i]=(c11[i]/c33[i]-1.)*0.5;
   del[i]=(c13[i]/c33[i]*c13[i]/c33[i]-1.)*0.5;
   v[i]=sqrt(c33[i]);
  }
  write("eps",eps,n);
  to_header("eps","n1",nx,"o1",ox,"d1",dx);
  to_header("eps","n2",nz,"o2",oz,"d2",dz);
  write("del",del,n);
  to_header("del","n1",nx,"o1",ox,"d1",dx);
  to_header("del","n2",nz,"o2",oz,"d2",dz);
  write("v",v,n);
  to_header("v","n1",nx,"o1",ox,"d1",dx);
  to_header("v","n2",nz,"o2",oz,"d2",dz);
 
  delete []eps;delete []del;delete []v;
 }

 if(direction==2){
  read("c11",c11,n); 
  read("c13",c13,n); 
  read("c33",c33,n); 
  float *vn=new float[n]();
  float *eta=new float[n]();
  float *del=new float[n]();
  #pragma omp parallel for default(shared)
  for(int i=0;i<n;++i){
   float vn2=c13[i]*c13[i]/c33[i];
   eta[i]=(c11[i]/vn2-1.)*0.5;
   del[i]=(vn2/c33[i]-1.)*0.5;
   vn[i]=sqrt(vn2);
  }
  write("eta",eta,n);
  to_header("eta","n1",nx,"o1",ox,"d1",dx);
  to_header("eta","n2",nz,"o2",oz,"d2",dz);
  write("del",del,n);
  to_header("del","n1",nx,"o1",ox,"d1",dx);
  to_header("del","n2",nz,"o2",oz,"d2",dz);
  write("vn",vn,n);
  to_header("vn","n1",nx,"o1",ox,"d1",dx);
  to_header("vn","n2",nz,"o2",oz,"d2",dz);
 
  delete []eta;delete []del;delete []vn;
 }

 if(direction==3){
  read("c11",c11,n); 
  read("c13",c13,n); 
  read("c33",c33,n); 
  float *vh=new float[n]();
  float *eps=new float[n]();
  float *eta=new float[n]();
  #pragma omp parallel for default(shared)
  for(int i=0;i<n;++i){
   eta[i]=(c11[i]*c33[i]/c13[i]/c13[i]-1.)*0.5;
   eps[i]=(c11[i]/c33[i]-1.)*0.5;
   vh[i]=sqrt(c11[i]);
  }
  write("eta",eta,n);
  to_header("eta","n1",nx,"o1",ox,"d1",dx);
  to_header("eta","n2",nz,"o2",oz,"d2",dz);
  write("eps",eps,n);
  to_header("eps","n1",nx,"o1",ox,"d1",dx);
  to_header("eps","n2",nz,"o2",oz,"d2",dz);
  write("vh",vh,n);
  to_header("vh","n1",nx,"o1",ox,"d1",dx);
  to_header("vh","n2",nz,"o2",oz,"d2",dz);
 
  delete []eta;delete []eps;delete []vh;
 }
 
 if(direction==10){
  float *v=new float[n](); read("v",v,n);
  float *eps=new float[n](); read("eps",eps,n);
  float *del=new float[n](); read("del",del,n);
  
  VEpsDel2Cij(c11,c13,c33,v,eps,del,1.,1.,1.,n);
  
  write("c11",c11,n);
  to_header("c11","n1",nx,"o1",ox,"d1",dx);
  to_header("c11","n2",nz,"o2",oz,"d2",dz);
  write("c13",c13,n);
  to_header("c13","n1",nx,"o1",ox,"d1",dx);
  to_header("c13","n2",nz,"o2",oz,"d2",dz);
  write("c33",c33,n);
  to_header("c33","n1",nx,"o1",ox,"d1",dx);
  to_header("c33","n2",nz,"o2",oz,"d2",dz);
  
  delete []eps;delete []del;delete []v;
 }

 if(direction==20){
  float *vn=new float[n](); read("vn",vn,n);
  float *eta=new float[n](); read("eta",eta,n);
  float *del=new float[n](); read("del",del,n);
  
  VnEtaDel2Cij(c11,c13,c33,vn,eta,del,1.,1.,1.,n); 
  
  write("c11",c11,n);
  to_header("c11","n1",nx,"o1",ox,"d1",dx);
  to_header("c11","n2",nz,"o2",oz,"d2",dz);
  write("c13",c13,n);
  to_header("c13","n1",nx,"o1",ox,"d1",dx);
  to_header("c13","n2",nz,"o2",oz,"d2",dz);
  write("c33",c33,n);
  to_header("c33","n1",nx,"o1",ox,"d1",dx);
  to_header("c33","n2",nz,"o2",oz,"d2",dz);
  
  delete []eta;delete []del;delete []vn;
 }

 if(direction==30){
  float *vh=new float[n](); read("vh",vh,n);
  float *eps=new float[n](); read("eps",eps,n);
  float *eta=new float[n](); read("eta",eta,n);
  
  VhEpsEta2Cij(c11,c13,c33,vh,eps,eta,1.,1.,1.,n);
  
  write("c11",c11,n);
  to_header("c11","n1",nx,"o1",ox,"d1",dx);
  to_header("c11","n2",nz,"o2",oz,"d2",dz);
  write("c13",c13,n);
  to_header("c13","n1",nx,"o1",ox,"d1",dx);
  to_header("c13","n2",nz,"o2",oz,"d2",dz);
  write("c33",c33,n);
  to_header("c33","n1",nx,"o1",ox,"d1",dx);
  to_header("c33","n2",nz,"o2",oz,"d2",dz);
  
  delete []eta;delete []eps;delete []vh;
 }

 if(direction==12){
  float *v=new float[n](); read("v",v,n);
  float *eps=new float[n](); read("eps",eps,n);
  float *del=new float[n](); read("del",del,n);
  float *vn=new float[n]();
  float *eta=new float[n]();
  #pragma omp parallel for default(shared)
  for(int i=0;i<n;++i){
   vn[i]=v[i]*sqrt(1.+2.*del[i]);
   eta[i]=(eps[i]-del[i])/(1.+2.*del[i]);
  }
  write("eta",eta,n);
  to_header("eta","n1",nx,"o1",ox,"d1",dx);
  to_header("eta","n2",nz,"o2",oz,"d2",dz);
  write("vn",vn,n);
  to_header("vn","n1",nx,"o1",ox,"d1",dx);
  to_header("vn","n2",nz,"o2",oz,"d2",dz);
 
  delete []eta;delete []vn;
 
  delete []eps;delete []del;delete []v;
 }

 if(direction==13){
  float *v=new float[n](); read("v",v,n);
  float *eps=new float[n](); read("eps",eps,n);
  float *del=new float[n](); read("del",del,n);
  float *vh=new float[n]();
  float *eta=new float[n]();
  #pragma omp parallel for default(shared)
  for(int i=0;i<n;++i){
   vh[i]=v[i]*sqrt(1.+2.*eps[i]);
   eta[i]=(eps[i]-del[i])/(1.+2.*del[i]);
  }
  write("eta",eta,n);
  to_header("eta","n1",nx,"o1",ox,"d1",dx);
  to_header("eta","n2",nz,"o2",oz,"d2",dz);
  write("vh",vh,n);
  to_header("vh","n1",nx,"o1",ox,"d1",dx);
  to_header("vh","n2",nz,"o2",oz,"d2",dz);
 
  delete []eta;delete []vh;
 
  delete []eps;delete []del;delete []v;
 }

 delete []c11;delete []c13;delete []c33;
 myio_close();

 return 0;
}
