#include <cstdio>
#include "myio.h"
#include "mylib.h"
#include "init.h"
#include "solvers.h"
#include "helix.h"
#include "steer.h"

using namespace std;

int main(int argc,char **argv){
 myio_init(argc,argv);

 int nnx,nnz,nx,nz,nt,npad;
 float ox,oz,ot,dx,dz,dt;
 init2d(nnx,nnz,nx,nz,nt,dx,dz,dt,ox,oz,ot,npad);

 float f; get_param("f",f);
 float *wavelet=new float[nt]();
 ricker(wavelet,f,nt,dt);
 
 int ns;
 float os,ds,zs;
 init_shot(ns,ds,os,zs);
 int *sloc=new int[2*ns]();
 init_shot_loc(sloc,ns,ds,os,zs,dx,dz,ox,oz,npad);
 
 int nr;
 float o_r,dr,zr;
 init_rec(nr,dr,o_r,zr);
 int *rloc=new int[2*nr]();
 init_rec_loc(rloc,nr,dr,o_r,zr,dx,dz,ox,oz,npad);
 
 float rate; get_param("rate",rate);
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 float *d0=new float[nr*nnt*ns]();
 read("data",d0,nr*nnt*ns);
 
 float *taper=new float[nnx*nnz]();
 init_abc(taper,nx,nz,npad);

 float wbottom; get_param("wbottom",wbottom);

 int niter; get_param("niter",niter);

 float a0; get_param("a0",a0);
 fprintf(stderr,"Starting step length %f\n",a0);
 
 int padded; get_param("padded",padded);
 
 int parameterize; get_param("parameterize",parameterize);

 if(parameterize==0){
  float *c11=new float[nnx*nnz]();
  float *c13=new float[nnx*nnz]();
  float *c33=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("c11",c11,nx,nz,npad); scale(c11,c11,1e-6,nnx*nnz);
   init_model("c13",c13,nx,nz,npad); scale(c13,c13,1e-6,nnx*nnz);
   init_model("c33",c33,nx,nz,npad); scale(c33,c33,1e-6,nnx*nnz);
  } 
  else{ 
   read("c11",c11,nnx*nnz);
   read("c13",c13,nnx*nnz);
   read("c33",c33,nnx*nnz);
  }
 
  int regularize; get_param("regularize",regularize);
  
  if(regularize){
   float alpha; get_param("alpha",alpha);
   
   float diptop; get_param("diptop",diptop);
   int ntop=npad+diptop/dz;
   int nfilters=1+(nnz-ntop)*nnx;
  
   int *na=new int[nfilters]();
   na[0]=8;
   
   int *patch=new int[nnx*nnz]();
   set(patch,0,ntop*nnx);
  
   float **aa=new float*[nfilters]();
   aa[0]=new float[na[0]]();
  
   int **lag=new int*[nfilters]();
   lag[0]=new int[na[0]]();
   
   float temp=lapfac(aa[0],lag[0],na[0],0.001,nnx,20);
   
   float *dip=new float[(nnz-ntop)*nnx]();
   read("dip",dip,(nnz-ntop)*nnx);
  
   for(int i=1;i<nfilters;++i){
    na[i]=3;
    aa[i]=new float[na[i]]();
    lag[i]=new int[na[i]]();
    steer0(aa[i],lag[i],na[i],dip[i-1],nnx); 
    patch[i-1+ntop*nnx]=i;
   }
  
   int precondition; get_param("precondition",precondition);
  
   if(precondition) pnlcg01(c11,c13,c33,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,alpha,aa,lag,na,patch);
   else nlcg01(c11,c13,c33,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,alpha,aa,lag,na,patch);
   
   delete []aa;delete []lag;
  } 
  else{
   nlcg0(c11,c13,c33,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom);
  }
  
  fprintf(stderr,"Done \n");
  delete []c11;delete []c13;delete []c33;
 }
 
 if(parameterize==1){
  float *v=new float[nnx*nnz]();
  float *eps=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("v",v,nx,nz,npad); 
   init_model("eps",eps,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("v",v,nnx*nnz);
   read("eps",eps,nnx*nnz);
   read("del",del,nnx*nnz);
  }
  
  float v0,eps0,del0;
  get_param("v0",v0,"eps0",eps0,"del0",del0);
 
  scale(v,v,1./v0,nnx*nnz);
  scale(eps,eps,1./eps0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
 
  int regularize; get_param("regularize",regularize);
  
  if(regularize){
   float alpha; get_param("alpha",alpha);
   
   float diptop; get_param("diptop",diptop);
   int ntop=npad+diptop/dz;
   int nfilters=1+(nnz-ntop)*nnx;
  
   int *na=new int[nfilters]();
   na[0]=8;
   
   int *patch=new int[nnx*nnz]();
   set(patch,0,ntop*nnx);
  
   float **aa=new float*[nfilters]();
   aa[0]=new float[na[0]]();
  
   int **lag=new int*[nfilters]();
   lag[0]=new int[na[0]]();
   
   float temp=lapfac(aa[0],lag[0],na[0],0.001,nnx,20);
   
   float *dip=new float[(nnz-ntop)*nnx]();
   read("dip",dip,(nnz-ntop)*nnx);
  
   for(int i=1;i<nfilters;++i){
    na[i]=3;
    aa[i]=new float[na[i]]();
    lag[i]=new int[na[i]]();
    steer0(aa[i],lag[i],na[i],dip[i-1],nnx); 
    patch[i-1+ntop*nnx]=i;
   }
  
   int precondition; get_param("precondition",precondition);
  
   if(precondition) pnlcg11(v,eps,del,v0,eps0,del0,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,alpha,aa,lag,na,patch);
   else nlcg11(v,eps,del,v0,eps0,del0,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,alpha,aa,lag,na,patch);
 
   for(int i=0;i<nfilters;++i){
    delete []aa[i];delete []lag[i];
   }
   delete []na;delete []patch; delete []aa;delete []lag;delete []dip;
  }
  else{
   nlcg1(v,eps,del,v0,eps0,del0,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom);
  }

  fprintf(stderr,"Done \n");
  delete []v;delete []eps;delete []del;
 }
 
 if(parameterize==2){
  float *vn=new float[nnx*nnz]();
  float *eta=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("vn",vn,nx,nz,npad); 
   init_model("eta",eta,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("vn",vn,nnx*nnz);
   read("eta",eta,nnx*nnz);
   read("del",del,nnx*nnz);
  }
  
  float vn0,eta0,del0;
  get_param("vn0",vn0,"eta0",eta0,"del0",del0);
 
  scale(vn,vn,1./vn0,nnx*nnz);
  scale(eta,eta,1./eta0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
 
  nlcg2(vn,eta,del,vn0,eta0,del0,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom);

  fprintf(stderr,"Done \n");
  delete []vn;delete []eta;delete []del;
 }
 
 if(parameterize==3){
  float *vh=new float[nnx*nnz]();
  float *eps=new float[nnx*nnz]();
  float *eta=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("vh",vh,nx,nz,npad); 
   init_model("eps",eps,nx,nz,npad); 
   init_model("eta",eta,nx,nz,npad); 
  } 
  else{ 
   read("vh",vh,nnx*nnz);
   read("eps",eps,nnx*nnz);
   read("eta",eta,nnx*nnz);
  }
  
  float vh0,eps0,eta0;
  get_param("vh0",vh0,"eps0",eps0,"eta0",eta0);
 
  scale(vh,vh,1./vh0,nnx*nnz);
  scale(eps,eps,1./eps0,nnx*nnz);
  scale(eta,eta,1./eta0,nnx*nnz);
 
  nlcg3(vh,eps,eta,vh0,eps0,eta0,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom);

  fprintf(stderr,"Done \n");
  delete []vh;delete []eps;delete []eta;
 }
 
 if(parameterize==4){
  float *v=new float[nnx*nnz]();
  float *vh=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("v",v,nx,nz,npad); 
   init_model("vh",vh,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("v",v,nnx*nnz);
   read("vh",vh,nnx*nnz);
   read("del",del,nnx*nnz);
  }
  
  float v0,vh0,del0;
  get_param("v0",v0,"vh0",vh0,"del0",del0);
 
  scale(v,v,1./v0,nnx*nnz);
  scale(vh,vh,1./vh0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
 
  nlcg4(v,vh,del,v0,vh0,del0,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom);

  fprintf(stderr,"Done \n");
  delete []v;delete []vh;delete []del;
 }
 
 if(parameterize==5){
  float *vn=new float[nnx*nnz]();
  float *vh=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("vn",vn,nx,nz,npad); 
   init_model("vh",vh,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("vn",vn,nnx*nnz);
   read("vh",vh,nnx*nnz);
   read("del",del,nnx*nnz);
  }
  
  float vn0,vh0,del0;
  get_param("vn0",vn0,"vh0",vh0,"del0",del0);
 
  scale(vn,vn,1./vn0,nnx*nnz);
  scale(vh,vh,1./vh0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
 
  nlcg5(vn,vh,del,vn0,vh0,del0,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom);

  fprintf(stderr,"Done \n");
  delete []vn;delete []vh;delete []del;
 }
 
 if(parameterize==6){
  float *v=new float[nnx*nnz]();
  float *vn=new float[nnx*nnz]();
  float *vh=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("v",v,nx,nz,npad); 
   init_model("vn",vn,nx,nz,npad); 
   init_model("vh",vh,nx,nz,npad); 
  } 
  else{ 
   read("v",v,nnx*nnz);
   read("vn",vn,nnx*nnz);
   read("vh",vh,nnx*nnz);
  }
  
  scale(v,v,1e-3,nnx*nnz);
  scale(vn,vn,1e-3,nnx*nnz);
  scale(vh,vh,1e-3,nnx*nnz);

  nlcg6(v,vn,vh,niter,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom);

  fprintf(stderr,"Done \n");
  delete []v;delete []vn;delete []vh;
 }
 
 delete []wavelet;delete []d0;delete []sloc;delete []rloc;delete []taper;

 myio_close();
 return 0;
}
