#include <cmath>
#include <omp.h>
#include <cstdio>

#include "myio.h"
#include "mylib.h"
#include "init.h"
#include "wave.h"
#include "helix.h"
#include "steer.h"

using namespace std;

int main(int argc,char **argc11){
 myio_init(argc,argc11);

int nnx,nnz,nx,nz,nt,npad;
 float ox,oz,ot,dx,dz,dt;
 init2d(nnx,nnz,nx,nz,nt,dx,dz,dt,ox,oz,ot,npad);

 float f; get_param("f",f);
 float *wavelet=new float[nt]();
 ricker(wavelet,f,nt,dt);

 int ns;
 float os,ds,zs;
 init_shot(ns,ds,os,zs);
 int *sloc=new int[2*ns]();
 init_shot_loc(sloc,ns,ds,os,zs,dx,dz,ox,oz,npad);
 
 int nr;
 float o_r,dr,zr;
 init_rec(nr,dr,o_r,zr);
 int *rloc=new int[2*nr]();
 init_rec_loc(rloc,nr,dr,o_r,zr,dx,dz,ox,oz,npad);
 
 float rate; get_param("rate",rate);
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 float *d0=new float[nr*nnt*ns]();
 read("data",d0,nr*nnt*ns);
 
 float *taper=new float[nnx*nnz]();
 init_abc(taper,nx,nz,npad);
 
 float wbottom; get_param("wbottom",wbottom);

 int padded; get_param("padded",padded);
 
 float *m=new float[nnx*nnz]();

 int parameterize; get_param("parameterize",parameterize);

 if(parameterize==0){
  float *c11=new float[nnx*nnz]();
  float *c13=new float[nnx*nnz]();
  float *c33=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("c11",c11,nx,nz,npad); scale(c11,c11,1e-6,nnx*nnz);
   init_model("c13",c13,nx,nz,npad); scale(c13,c13,1e-6,nnx*nnz);
   init_model("c33",c33,nx,nz,npad); scale(c33,c33,1e-6,nnx*nnz);
  } 
  else{ 
   read("c11",c11,nnx*nnz);
   read("c13",c13,nnx*nnz);
   read("c33",c33,nnx*nnz);
  }
 
  float *gc11=new float[nnx*nnz]();
  float *gc13=new float[nnx*nnz]();
  float *gc33=new float[nnx*nnz]();
  
  int regularize; get_param("regularize",regularize);
  
  if(regularize){
   float alpha; get_param("alpha",alpha);
   
   float diptop; get_param("diptop",diptop);
   int ntop=npad+diptop/dz;
   int nfilters=1+(nnz-ntop)*nnx;
  
   int *na=new int[nfilters]();
   na[0]=8;
   
   int *patch=new int[nnx*nnz]();
   set(patch,0,ntop*nnx);
  
   float **aa=new float*[nfilters]();
   aa[0]=new float[na[0]]();
  
   int **lag=new int*[nfilters]();
   lag[0]=new int[na[0]]();
   
   float temp=lapfac(aa[0],lag[0],na[0],0.001,nnx,20);
   
   float *dip=new float[(nnz-ntop)*nnx]();
   read("dip",dip,(nnz-ntop)*nnx);
  
   for(int i=1;i<nfilters;++i){
    na[i]=3;
    aa[i]=new float[na[i]]();
    lag[i]=new int[na[i]]();
    steer0(aa[i],lag[i],na[i],dip[i-1],nnx); 
    patch[i-1+ntop*nnx]=i;
   }
  
   float datafit,reg;
   float val=objFuncGradient01(datafit,reg,gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   cout<<"Datafit "<<datafit<<" reg "<<reg<<" objfunc "<<val<<endl;
   
   int precondition; get_param("precondition",precondition);
   
   if(precondition){
    float *tc11=new float[nnx*nnz](); ns_helixdecon(gc11,tc11,nnx*nnz,aa,lag,na,patch,true);
    float *tc13=new float[nnx*nnz](); ns_helixdecon(gc13,tc13,nnx*nnz,aa,lag,na,patch,true);
    float *tc33=new float[nnx*nnz](); ns_helixdecon(gc33,tc33,nnx*nnz,aa,lag,na,patch,true);
    float *ttc11=new float[nnx*nnz](); ns_helixdecon(ttc11,tc11,nnx*nnz,aa,lag,na,patch,false);
    float *ttc13=new float[nnx*nnz](); ns_helixdecon(ttc13,tc13,nnx*nnz,aa,lag,na,patch,false);
    float *ttc33=new float[nnx*nnz](); ns_helixdecon(ttc33,tc33,nnx*nnz,aa,lag,na,patch,false);
    
    int n=npad+wbottom/dz+1;
    memset(ttc11,0,n*nnx*sizeof(float)); 
    memset(ttc13,0,n*nnx*sizeof(float)); 
    memset(ttc33,0,n*nnx*sizeof(float)); 
    
	write("pgc11",ttc11,nnx*nnz);
    to_header("pgc11","n1",nnx,"o1",-dx*npad,"d1",dx);
    to_header("pgc11","n2",nnz,"o2",-dz*npad,"d2",dz);
    write("pgc13",ttc13,nnx*nnz);
    to_header("pgc13","n1",nnx,"o1",-dx*npad,"d1",dx);
    to_header("pgc13","n2",nnz,"o2",-dz*npad,"d2",dz);
    write("pgc33",ttc33,nnx*nnz);
    to_header("pgc33","n1",nnx,"o1",-dx*npad,"d1",dx);
    to_header("pgc33","n2",nnz,"o2",-dz*npad,"d2",dz);
    
	delete []tc11;delete []tc13;delete []tc33;
    delete []ttc11;delete []ttc13;delete []ttc33;
   } 
   
   for(int i=0;i<nfilters;++i){
    delete []aa[i];delete []lag[i];
   }
   delete []na;delete []patch; delete []aa;delete []lag;delete []dip;
  }
  else{
   float val=objFuncGradient0(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   cout<<"Objfunc "<<val<<endl;
  }
  
  write("gc11",gc11,nnx*nnz);
  to_header("gc11","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gc11","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("gc13",gc13,nnx*nnz);
  to_header("gc13","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gc13","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("gc33",gc33,nnx*nnz);
  to_header("gc33","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gc33","n2",nnz,"o2",-dz*npad,"d2",dz);
  
  delete []c11;delete []c13;delete []c33;
  delete []gc11;delete []gc13;delete []gc33;
 } 

 if(parameterize==1){
  float *v=new float[nnx*nnz]();
  float *eps=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("v",v,nx,nz,npad); 
   init_model("eps",eps,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("v",v,nnx*nnz);
   read("eps",eps,nnx*nnz);
   read("del",del,nnx*nnz);
  }
 
  float v0,eps0,del0;
  get_param("v0",v0,"eps0",eps0,"del0",del0);
 
  scale(v,v,1./v0,nnx*nnz);
  scale(eps,eps,1./eps0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
  
  float *gv=new float[nnx*nnz]();
  float *geps=new float[nnx*nnz]();
  float *gdel=new float[nnx*nnz]();
  
  int regularize; get_param("regularize",regularize);
  
  if(regularize){
   float alpha; get_param("alpha",alpha);
   
   float diptop; get_param("diptop",diptop);
   int ntop=npad+diptop/dz;
   int nfilters=1+(nnz-ntop)*nnx;
  
   int *na=new int[nfilters]();
   na[0]=8;
   
   int *patch=new int[nnx*nnz]();
   set(patch,0,ntop*nnx);
  
   float **aa=new float*[nfilters]();
   aa[0]=new float[na[0]]();
  
   int **lag=new int*[nfilters]();
   lag[0]=new int[na[0]]();
   
   float temp=lapfac(aa[0],lag[0],na[0],0.001,nnx,20);
   
   float *dip=new float[(nnz-ntop)*nnx]();
   read("dip",dip,(nnz-ntop)*nnx);
  
   for(int i=1;i<nfilters;++i){
    na[i]=3;
    aa[i]=new float[na[i]]();
    lag[i]=new int[na[i]]();
    steer0(aa[i],lag[i],na[i],dip[i-1],nnx); 
    patch[i-1+ntop*nnx]=i;
   }
  
   float datafit,reg;
   float val=objFuncGradient11(datafit,reg,gv,geps,gdel,d0,v,eps,del,v0,eps0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   cout<<"Datafit "<<datafit<<" reg "<<reg<<" objfunc "<<val<<endl;
   
   int precondition; get_param("precondition",precondition);
   
   if(precondition){
    float *tv=new float[nnx*nnz](); ns_helixdecon(gv,tv,nnx*nnz,aa,lag,na,patch,true);
    float *teps=new float[nnx*nnz](); ns_helixdecon(geps,teps,nnx*nnz,aa,lag,na,patch,true);
    float *tdel=new float[nnx*nnz](); ns_helixdecon(gdel,tdel,nnx*nnz,aa,lag,na,patch,true);
    float *ttv=new float[nnx*nnz](); ns_helixdecon(ttv,tv,nnx*nnz,aa,lag,na,patch,false);
    float *tteps=new float[nnx*nnz](); ns_helixdecon(tteps,teps,nnx*nnz,aa,lag,na,patch,false);
    float *ttdel=new float[nnx*nnz](); ns_helixdecon(ttdel,tdel,nnx*nnz,aa,lag,na,patch,false);
    
	int n=npad+wbottom/dz+1;
    memset(ttv,0,n*nnx*sizeof(float)); 
    memset(tteps,0,n*nnx*sizeof(float)); 
    memset(ttdel,0,n*nnx*sizeof(float)); 
    
	write("pgv",ttv,nnx*nnz);
    to_header("pgv","n1",nnx,"o1",-dx*npad,"d1",dx);
    to_header("pgv","n2",nnz,"o2",-dz*npad,"d2",dz);
    write("pgeps",tteps,nnx*nnz);
    to_header("pgeps","n1",nnx,"o1",-dx*npad,"d1",dx);
    to_header("pgeps","n2",nnz,"o2",-dz*npad,"d2",dz);
    write("pgdel",ttdel,nnx*nnz);
    to_header("pgdel","n1",nnx,"o1",-dx*npad,"d1",dx);
    to_header("pgdel","n2",nnz,"o2",-dz*npad,"d2",dz);

	delete []tv;delete []teps;delete []tdel;
    delete []ttv;delete []tteps;delete []ttdel;
   } 
    
   for(int i=0;i<nfilters;++i){
    delete []aa[i];delete []lag[i];
   }
   delete []na;delete []patch; delete []aa;delete []lag;delete []dip;
  }
  else{
   float val=objFuncGradient1(gv,geps,gdel,d0,v,eps,del,v0,eps0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   cout<<"Objfunc "<<val<<endl;
  }
 
  write("gv",gv,nnx*nnz);
  to_header("gv","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gv","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("geps",geps,nnx*nnz);
  to_header("geps","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("geps","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("gdel",gdel,nnx*nnz);
  to_header("gdel","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gdel","n2",nnz,"o2",-dz*npad,"d2",dz);
 
  delete []v; delete []eps; delete []del;
  delete []gv; delete []geps; delete []gdel;
 }

 if(parameterize==2){
  float *vn=new float[nnx*nnz]();
  float *eta=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("vn",vn,nx,nz,npad); 
   init_model("eta",eta,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("vn",vn,nnx*nnz);
   read("eta",eta,nnx*nnz);
   read("del",del,nnx*nnz);
  }
 
  float vn0,eta0,del0;
  get_param("vn0",vn0,"eta0",eta0,"del0",del0);
 
  scale(vn,vn,1./vn0,nnx*nnz);
  scale(eta,eta,1./eta0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
  
  float *gvn=new float[nnx*nnz]();
  float *geta=new float[nnx*nnz]();
  float *gdel=new float[nnx*nnz]();
  
  float val=objFuncGradient2(gvn,geta,gdel,d0,vn,eta,del,vn0,eta0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  cout<<"Objfunc "<<val<<endl;
 
  write("gvn",gvn,nnx*nnz);
  to_header("gvn","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gvn","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("geta",geta,nnx*nnz);
  to_header("geta","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("geta","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("gdel",gdel,nnx*nnz);
  to_header("gdel","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gdel","n2",nnz,"o2",-dz*npad,"d2",dz);
 
  delete []vn; delete []eta; delete []del;
  delete []gvn; delete []geta; delete []gdel;
 }
 
 if(parameterize==3){
  float *vh=new float[nnx*nnz]();
  float *eps=new float[nnx*nnz]();
  float *eta=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("vh",vh,nx,nz,npad); 
   init_model("eps",eps,nx,nz,npad); 
   init_model("eta",eta,nx,nz,npad); 
  } 
  else{ 
   read("vh",vh,nnx*nnz);
   read("eps",eps,nnx*nnz);
   read("eta",eta,nnx*nnz);
  }
 
  float vh0,eps0,eta0;
  get_param("vh0",vh0,"eps0",eps0,"eta0",eta0);
 
  scale(vh,vh,1./vh0,nnx*nnz);
  scale(eps,eps,1./eps0,nnx*nnz);
  scale(eta,eta,1./eta0,nnx*nnz);
  
  float *gvh=new float[nnx*nnz]();
  float *geps=new float[nnx*nnz]();
  float *geta=new float[nnx*nnz]();
  
  float val=objFuncGradient3(gvh,geps,geta,d0,vh,eps,eta,vh0,eps0,eta0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  cout<<"Objfunc "<<val<<endl;
 
  write("gvh",gvh,nnx*nnz);
  to_header("gvh","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gvh","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("geps",geps,nnx*nnz);
  to_header("geps","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("geps","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("geta",geta,nnx*nnz);
  to_header("geta","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("geta","n2",nnz,"o2",-dz*npad,"d2",dz);
 
  delete []vh; delete []eps; delete []eta;
  delete []gvh; delete []geps; delete []geta;
 }
 
 if(parameterize==4){
  float *v=new float[nnx*nnz]();
  float *vh=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("v",v,nx,nz,npad); 
   init_model("vh",vh,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("v",v,nnx*nnz);
   read("vh",vh,nnx*nnz);
   read("del",del,nnx*nnz);
  }
  
  float v0,vh0,del0;
  get_param("v0",v0,"vh0",vh0,"del0",del0);
 
  scale(v,v,1./v0,nnx*nnz);
  scale(vh,vh,1./vh0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
  
  float *gv=new float[nnx*nnz]();
  float *gvh=new float[nnx*nnz]();
  float *gdel=new float[nnx*nnz]();
  
  float val=objFuncGradient4(gv,gvh,gdel,d0,v,vh,del,v0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  cout<<"Objfunc "<<val<<endl;
 
  write("gv",gv,nnx*nnz);
  to_header("gv","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gv","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("gvh",gvh,nnx*nnz);
  to_header("gvh","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gvh","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("gdel",gdel,nnx*nnz);
  to_header("gdel","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gdel","n2",nnz,"o2",-dz*npad,"d2",dz);
 
  delete []v; delete []vh; delete []del;
  delete []gv; delete []gvh; delete []gdel;
 }
 
 if(parameterize==5){
  float *vn=new float[nnx*nnz]();
  float *vh=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("vn",vn,nx,nz,npad); 
   init_model("vh",vh,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("vn",vn,nnx*nnz);
   read("vh",vh,nnx*nnz);
   read("del",del,nnx*nnz);
  }
  
  float vn0,vh0,del0;
  get_param("vn0",vn0,"vh0",vh0,"del0",del0);
 
  scale(vn,vn,1./vn0,nnx*nnz);
  scale(vh,vh,1./vh0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
  
  float *gvn=new float[nnx*nnz]();
  float *gvh=new float[nnx*nnz]();
  float *gdel=new float[nnx*nnz]();
  
  float val=objFuncGradient5(gvn,gvh,gdel,d0,vn,vh,del,vn0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  cout<<"Objfunc "<<val<<endl;
 
  write("gvn",gvn,nnx*nnz);
  to_header("gvn","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gvn","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("gvh",gvh,nnx*nnz);
  to_header("gvh","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gvh","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("gdel",gdel,nnx*nnz);
  to_header("gdel","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gdel","n2",nnz,"o2",-dz*npad,"d2",dz);
 
  delete []vn; delete []vh; delete []del;
  delete []gvn; delete []gvh; delete []gdel;
 }
 
 if(parameterize==6){
  float *v=new float[nnx*nnz]();
  float *vn=new float[nnx*nnz]();
  float *vh=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("v",v,nx,nz,npad); 
   init_model("vn",vn,nx,nz,npad); 
   init_model("vh",vh,nx,nz,npad); 
  } 
  else{ 
   read("v",v,nnx*nnz);
   read("vn",vn,nnx*nnz);
   read("vh",vh,nnx*nnz);
  }
  
  scale(v,v,1e-3,nnx*nnz);
  scale(vn,vn,1e-3,nnx*nnz);
  scale(vh,vh,1e-3,nnx*nnz);

  float *gv=new float[nnx*nnz]();
  float *gvn=new float[nnx*nnz]();
  float *gvh=new float[nnx*nnz]();
  
  float val=objFuncGradient6(gv,gvn,gvh,d0,v,vn,vh,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  cout<<"Objfunc "<<val<<endl;
 
  write("gv",gv,nnx*nnz);
  to_header("gv","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gv","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("gvn",gvn,nnx*nnz);
  to_header("gvn","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gvn","n2",nnz,"o2",-dz*npad,"d2",dz);
  write("gvh",gvh,nnx*nnz);
  to_header("gvh","n1",nnx,"o1",-dx*npad,"d1",dx);
  to_header("gvh","n2",nnz,"o2",-dz*npad,"d2",dz);
 
  delete []vn; delete []vh; delete []v;
  delete []gvn; delete []gvh; delete []gv;
 }
 
 delete []wavelet;delete []d0;delete []sloc;delete []rloc;delete []taper;
 delete []m;

 myio_close();
 return 0;
}
