#include <math.h>
#include <string.h>
#include <iostream>
#include "helix.h"

void helixcon(float *y,float *x,int n,float *a,int *lag,int na,bool adj){
 if(adj){
  memcpy(x,y,sizeof(float)*n);
  for(int ia=0;ia<na;++ia){
   for(int iy=lag[ia];iy<n;++iy){
    int ix=iy-lag[ia];
	x[ix]+=y[iy]*a[ia];
   }
  }
 }else{
  memcpy(y,x,sizeof(float)*n);
  for(int ia=0;ia<na;++ia){
   for(int iy=lag[ia];iy<n;++iy){
    int ix=iy-lag[ia];
	y[iy]+=x[ix]*a[ia];
   }
  }
 }
 return;
}

void helixdecon0(float *y,float *x,int n,float *a,int *lag,int na,bool adj){
 float *tt=new float[n]();
 if(adj){
  for(int ix=n-1;ix>=0;--ix){
   tt[ix]=y[ix];
   for(int ia=0;ia<na;++ia){
    int iy=ix+lag[ia];
	if(iy>=n) continue;
	tt[ix]-=tt[iy]*a[ia];
   }
  }
  memcpy(x,tt,sizeof(float)*n);
 }else{
  for(int iy=0;iy<n;++iy){
   tt[iy]=x[iy];
   for(int ia=0;ia<na;++ia){
    int ix=iy-lag[ia];
	if(ix<0) continue;
	tt[iy]-=tt[ix]*a[ia];
   }
  }
  memcpy(y,tt,sizeof(float)*n);
 }
 delete []tt;
 return;
}

void helixdecon(float *y,float *x,int n,float *a,int *lag,int na,bool adj){
 if(adj){
  for(int ix=n-1;ix>=0;--ix){
   float tt=y[ix];
   for(int ia=0;ia<na;++ia){
    int iy=ix+lag[ia];
	if(iy>=n) continue;
	tt-=x[iy]*a[ia];
   }
   x[ix]=tt;
  }
 }else{
  for(int iy=0;iy<n;++iy){
   float tt =x[iy];
   for(int ia=0;ia<na;++ia){
    int ix=iy-lag[ia];
	if(ix<0) continue;
	tt-=y[ix]*a[ia];
   }
   y[iy]=tt;
  }
 }
 return;
}

float lapfac(float *a,int *lag,int na,float eps,int n1,int niter){
 float *lap=new float[2]();
 int *laglap=new int[2]();
 float lap0=4.+eps;
 laglap[0]=1;laglap[1]=n1;
 lap[0]=-1.;lap[1]=-1.;
 for(int ia=0;ia<na/2;++ia){
  lag[ia]=ia+1;
  lag[ia+na/2]=n1+ia-na/2+1;
 }
 float a0=wilson_factor(a,lag,na,10*n1,niter,lap0,lap,laglap,2);
 delete []lap;delete []laglap;
 return a0;
}

float wilson_factor(float *a,int *lag,int na,int nmax,int niter,float s0,const float *ss,const int *lags,int ns){
 int n=nmax,n2=2*n-1;
 float *au=new float[n2]();
 float *bb=new float[n2]();
 float *cc=new float[n2]();
 float *b=new float[n]();
 float *c=new float[n]();
 au[n-1]=s0;
 b[0]=1.;
 for(int is=0;is<ns;++is){
  au[n-1+lags[is]]=ss[is];
  au[n-1-lags[is]]=ss[is];
 }
 for(int iter=0;iter<niter;++iter){
  helixdecon(bb,au,n2,a,lag,na,false);
  helixdecon(bb,cc,n2,a,lag,na,true);
  float eps=0.;
  for(int i=1;i<n;++i){
   b[i]=0.5*(cc[n-1+i]+cc[n-1-i])/cc[n-1];
   if(fabs(b[i])>eps) eps=fabs(b[i]);
  }
  if(eps<1e-8) break;
  helixcon(c,b,n,a,lag,na,false);
  for(int i=0;i<na;++i) a[i]=c[lag[i]];
 }
 float a0=sqrt(cc[n-1]);
 delete []au;delete []bb;delete []cc;delete []b;delete []c;
 return a0;
}

void ns_helixcon(float *y,float *x,int n,float **a,int **lag,int *na,int *patch,bool adj){
 if(adj){
  memcpy(x,y,sizeof(float)*n);
  for(int iy=0;iy<n;++iy){
   int i=patch[iy];
   for(int ia=0;ia<na[i];++ia){
    int ix=iy-lag[i][ia];
	if(ix<0) continue;
	x[ix]+=y[iy]*a[i][ia];
   }
  }
 }else{
  memcpy(y,x,sizeof(float)*n);
  for(int iy=0;iy<n;++iy){
   int i=patch[iy];
   for(int ia=0;ia<na[i];++ia){
    int ix=iy-lag[i][ia];
	if(ix<0) continue;
	y[iy]+=x[ix]*a[i][ia];
   }
  }
 }
 return;
}

void ns_helixdecon(float *y,float *x,int n,float **a,int **lag,int *na,int *patch,bool adj){
 if(adj){
  float *tt=new float[n]();
  memcpy(tt,y,n*sizeof(float));
  for(int iy=n-1;iy>=0;--iy){
   int i=patch[iy];
   for(int ia=0;ia<na[i];++ia){
    int ix=iy-lag[i][ia];
	if(ix<0) continue;
	tt[ix]-=tt[iy]*a[i][ia];
   }
  }
  memcpy(x,tt,n*sizeof(float));
  delete []tt;
 }else{
  for(int iy=0;iy<n;++iy){
   int i=patch[iy];
   float tt=x[iy];
   for(int ia=0;ia<na[i];++ia){
    int ix=iy-lag[i][ia];
	if(ix<0) continue;
	tt-=y[ix]*a[i][ia];
   }
   y[iy]=tt;
  }
 }
 return;
}

