#include <cmath>
#include <omp.h>
#include <cstdio>
#include <iostream>
#include "myio.h"
#include "mylib.h"
#include "init.h"
#include "wave.h"
#include "helix.h"
#include "steer.h"

using namespace std;

int main(int argc,char **arpv){
 myio_init(argc,arpv);

 int nnx,nnz,nx,nz,nt,npad;
 float ox,oz,ot,dx,dz,dt;
 init2d(nnx,nnz,nx,nz,nt,dx,dz,dt,ox,oz,ot,npad);

 float f; get_param("f",f);
 float *wavelet=new float[nt]();
 ricker(wavelet,f,nt,dt);
 
 int ns;
 float os,ds,zs;
 init_shot(ns,ds,os,zs);
 int *sloc=new int[2*ns]();
 init_shot_loc(sloc,ns,ds,os,zs,dx,dz,ox,oz,npad);
 
 int nr;
 float o_r,dr,zr;
 init_rec(nr,dr,o_r,zr);
 int *rloc=new int[2*nr]();
 init_rec_loc(rloc,nr,dr,o_r,zr,dx,dz,ox,oz,npad);
 
 float rate; get_param("rate",rate);
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 float *d0=new float[nr*nnt*ns]();
 read("data",d0,nr*nnt*ns);
 
 float *taper=new float[nnx*nnz]();
 init_abc(taper,nx,nz,npad);
 
 int padded; get_param("padded",padded);
 
 float *m=new float[nnx*nnz]();
 
 float s0,a,b;
 get_param("s0",s0,"a",a,"b",b);

 int parameterize; get_param("parameterize",parameterize);

 if(parameterize==0){
  float *c11=new float[nnx*nnz]();
  float *c13=new float[nnx*nnz]();
  float *c33=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("c11",c11,nx,nz,npad); scale(c11,c11,1e-6,nnx*nnz);
   init_model("c13",c13,nx,nz,npad); scale(c13,c13,1e-6,nnx*nnz);
   init_model("c33",c33,nx,nz,npad); scale(c33,c33,1e-6,nnx*nnz);
  } 
  else{ 
   read("c11",c11,nnx*nnz);
   read("c13",c13,nnx*nnz);
   read("c33",c33,nnx*nnz);
  }
  
  float val=objFunc0(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
  cout<<"initial obj func "<<val<<endl;
  
  float *pc11=new float[nnx*nnz](); read("pc11",pc11,nnx*nnz);
  float *pc13=new float[nnx*nnz](); read("pc13",pc13,nnx*nnz);
  float *pc33=new float[nnx*nnz](); read("pc33",pc33,nnx*nnz);
 
  float *tc11=new float[nnx*nnz]();
  float *tc13=new float[nnx*nnz]();
  float *tc33=new float[nnx*nnz]();
  
  float s=(-1.)*s0;
  for(int i=1;i<10;++i){
   lin_comb(tc11,1.,c11,s,pc11,nnx*nnz);
   lin_comb(tc13,1.,c13,s,pc13,nnx*nnz);
   lin_comb(tc33,1.,c33,s,pc33,nnx*nnz);
  
   val=objFunc0(d0,tc11,tc13,tc33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
   cout<<"Step "<<s<<" Obj func "<<val<<endl;
   s=a*s+b;
  }
  
  delete []c11;delete []c13;delete []c33;
  delete []pc11;delete []pc13;delete []pc33;
  delete []tc11;delete []tc13;delete []tc33;
 }  
 
 if(parameterize==1){
  float *v=new float[nnx*nnz]();
  float *eps=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("v",v,nx,nz,npad); 
   init_model("eps",eps,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("v",v,nnx*nnz);
   read("eps",eps,nnx*nnz);
   read("del",del,nnx*nnz);
  }
 
  float v0,eps0,del0;
  get_param("v0",v0,"eps0",eps0,"del0",del0);
  
  scale(v,v,1./v0,nnx*nnz);
  scale(eps,eps,1./eps0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
  
  float *pv=new float[nnx*nnz](); read("pv",pv,nnx*nnz); 
  float *peps=new float[nnx*nnz](); read("peps",peps,nnx*nnz); 
  float *pdel=new float[nnx*nnz](); read("pdel",pdel,nnx*nnz); 
 
  float *tv=new float[nnx*nnz]();
  float *teps=new float[nnx*nnz]();
  float *tdel=new float[nnx*nnz]();
  
  int regularize; get_param("regularize",regularize);

  float val;
  if(regularize){
   float alpha; get_param("alpha",alpha);
   
   float diptop; get_param("diptop",diptop);
   int ntop=npad+diptop/dz;
   int nfilters=1+(nnz-ntop)*nnx;
  
   int *na=new int[nfilters]();
   na[0]=8;
   
   int *patch=new int[nnx*nnz]();
   set(patch,0,ntop*nnx);
  
   float **aa=new float*[nfilters]();
   aa[0]=new float[na[0]]();
  
   int **lag=new int*[nfilters]();
   lag[0]=new int[na[0]]();
   
   float temp=lapfac(aa[0],lag[0],na[0],0.001,nnx,20);
   
   float *dip=new float[(nnz-ntop)*nnx]();
   read("dip",dip,(nnz-ntop)*nnx);
  
   for(int i=1;i<nfilters;++i){
    na[i]=3;
    aa[i]=new float[na[i]]();
    lag[i]=new int[na[i]]();
    steer0(aa[i],lag[i],na[i],dip[i-1],nnx); 
    patch[i-1+ntop*nnx]=i;
   }
  
   float datafit,reg;
   val=objFunc11(datafit,reg,d0,v,eps,del,v0,eps0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m,alpha,aa,lag,na,patch);
   cout<<"Initial datafit "<<datafit<<" reg "<<reg<<" objfunc "<<val<<endl;
   
   float s=(-1.)*s0;
   for(int i=1;i<10;++i){
    
    lin_comb(tv,1.,v,s,pv,nnx*nnz);
    lin_comb(teps,1.,eps,s,peps,nnx*nnz);
    lin_comb(tdel,1.,del,s,pdel,nnx*nnz);
   
    val=objFunc11(datafit,reg,d0,tv,teps,tdel,v0,eps0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m,alpha,aa,lag,na,patch);
    
    cout<<"Step "<<s<<" Obj func "<<val<<" datafit "<<datafit<<" reg "<<reg<<endl;
    s=a*s+b;
   }
   
   for(int i=0;i<nfilters;++i){
    delete []aa[i];delete []lag[i];
   }
   delete []na;delete []patch; delete []aa;delete []lag;delete []dip;
  }
  else{
   val=objFunc1(d0,v,eps,del,v0,eps0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
   cout<<"initial obj func "<<val<<endl;
   
   float s=(-1.)*s0;
   for(int i=1;i<10;++i){
    
    lin_comb(tv,1.,v,s,pv,nnx*nnz);
    lin_comb(teps,1.,eps,s,peps,nnx*nnz);
    lin_comb(tdel,1.,del,s,pdel,nnx*nnz);
   
    val=objFunc1(d0,tv,teps,tdel,v0,eps0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
    
    cout<<"Step "<<s<<" Obj func "<<val<<endl;
    s=a*s+b;
   }
  }
  
  delete []v;delete []eps;delete []del;
  delete []pv;delete []peps;delete []pdel;
  delete []tv;delete []teps;delete []tdel;
 }
 
 if(parameterize==2){
  float *vn=new float[nnx*nnz]();
  float *eta=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("vn",vn,nx,nz,npad); 
   init_model("eta",eta,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("vn",vn,nnx*nnz);
   read("eta",eta,nnx*nnz);
   read("del",del,nnx*nnz);
  }
 
  float vn0,eta0,del0;
  get_param("vn0",vn0,"eta0",eta0,"del0",del0);
  
  scale(vn,vn,1./vn0,nnx*nnz);
  scale(eta,eta,1./eta0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
  
  float val=objFunc2(d0,vn,eta,del,vn0,eta0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
  cout<<"initial obj func "<<val<<endl;
  
  float *pvn=new float[nnx*nnz](); read("pvn",pvn,nnx*nnz); 
  float *peta=new float[nnx*nnz](); read("peta",peta,nnx*nnz); 
  float *pdel=new float[nnx*nnz](); read("pdel",pdel,nnx*nnz); 
 
  float *tvn=new float[nnx*nnz]();
  float *teta=new float[nnx*nnz]();
  float *tdel=new float[nnx*nnz]();
 
  float s=(-1.)*s0;
  for(int i=1;i<10;++i){
   
   lin_comb(tvn,1.,vn,s,pvn,nnx*nnz);
   lin_comb(teta,1.,eta,s,peta,nnx*nnz);
   lin_comb(tdel,1.,del,s,pdel,nnx*nnz);
  
   val=objFunc2(d0,tvn,teta,tdel,vn0,eta0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
   
   cout<<"Step "<<s<<" Obj func "<<val<<endl;
   s=a*s+b;
  }
  
  delete []vn;delete []eta;delete []del;
  delete []pvn;delete []peta;delete []pdel;
  delete []tvn;delete []teta;delete []tdel;
 }
 
 if(parameterize==3){
  float *vh=new float[nnx*nnz]();
  float *eps=new float[nnx*nnz]();
  float *eta=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("vh",vh,nx,nz,npad); 
   init_model("eps",eps,nx,nz,npad); 
   init_model("eta",eta,nx,nz,npad); 
  } 
  else{ 
   read("vh",vh,nnx*nnz);
   read("eps",eps,nnx*nnz);
   read("eta",eta,nnx*nnz);
  }
 
  float vh0,eps0,eta0;
  get_param("vh0",vh0,"eps0",eps0,"eta0",eta0);
  
  scale(vh,vh,1./vh0,nnx*nnz);
  scale(eps,eps,1./eps0,nnx*nnz);
  scale(eta,eta,1./eta0,nnx*nnz);
  
  float val=objFunc3(d0,vh,eps,eta,vh0,eps0,eta0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
  cout<<"initial obj func "<<val<<endl;
  
  float *pvh=new float[nnx*nnz](); read("pvh",pvh,nnx*nnz); 
  float *peps=new float[nnx*nnz](); read("peps",peps,nnx*nnz); 
  float *peta=new float[nnx*nnz](); read("peta",peta,nnx*nnz); 
 
  float *tvh=new float[nnx*nnz]();
  float *teps=new float[nnx*nnz]();
  float *teta=new float[nnx*nnz]();
 
  float s=(-1.)*s0;
  for(int i=1;i<10;++i){
   
   lin_comb(tvh,1.,vh,s,pvh,nnx*nnz);
   lin_comb(teps,1.,eps,s,peps,nnx*nnz);
   lin_comb(teta,1.,eta,s,peta,nnx*nnz);
  
   val=objFunc3(d0,tvh,teps,teta,vh0,eps0,eta0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
   
   cout<<"Step "<<s<<" Obj func "<<val<<endl;
   s=a*s+b;
  }
  
  delete []vh;delete []eps;delete []eta;
  delete []pvh;delete []peps;delete []peta;
  delete []tvh;delete []teps;delete []teta;
 }
 
 if(parameterize==4){
  float *v=new float[nnx*nnz]();
  float *vh=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("v",v,nx,nz,npad); 
   init_model("vh",vh,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("v",v,nnx*nnz);
   read("vh",vh,nnx*nnz);
   read("del",del,nnx*nnz);
  }
 
  float v0,vh0,del0;
  get_param("v0",v0,"vh0",vh0,"del0",del0);
  
  scale(v,v,1./v0,nnx*nnz);
  scale(vh,vh,1./vh0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
  
  float val=objFunc4(d0,v,vh,del,v0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
  cout<<"initial obj func "<<val<<endl;
  
  float *pv=new float[nnx*nnz](); read("pv",pv,nnx*nnz); 
  float *pvh=new float[nnx*nnz](); read("pvh",pvh,nnx*nnz); 
  float *pdel=new float[nnx*nnz](); read("pdel",pdel,nnx*nnz); 
 
  float *tv=new float[nnx*nnz]();
  float *tvh=new float[nnx*nnz]();
  float *tdel=new float[nnx*nnz]();
 
  float s=(-1.)*s0;
  for(int i=1;i<10;++i){
   
   lin_comb(tv,1.,v,s,pv,nnx*nnz);
   lin_comb(tvh,1.,vh,s,pvh,nnx*nnz);
   lin_comb(tdel,1.,del,s,pdel,nnx*nnz);
  
   val=objFunc4(d0,tv,tvh,tdel,v0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
   
   cout<<"Step "<<s<<" Obj func "<<val<<endl;
   s=a*s+b;
  }
  
  delete []v;delete []vh;delete []del;
  delete []pv;delete []pvh;delete []pdel;
  delete []tv;delete []tvh;delete []tdel;
 }
 
 if(parameterize==5){
  float *vn=new float[nnx*nnz]();
  float *vh=new float[nnx*nnz]();
  float *del=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("vn",vn,nx,nz,npad); 
   init_model("vh",vh,nx,nz,npad); 
   init_model("del",del,nx,nz,npad); 
  } 
  else{ 
   read("vn",vn,nnx*nnz);
   read("vh",vh,nnx*nnz);
   read("del",del,nnx*nnz);
  }
 
  float vn0,vh0,del0;
  get_param("vn0",vn0,"vh0",vh0,"del0",del0);
  
  scale(vn,vn,1./vn0,nnx*nnz);
  scale(vh,vh,1./vh0,nnx*nnz);
  scale(del,del,1./del0,nnx*nnz);
  
  float val=objFunc5(d0,vn,vh,del,vn0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
  cout<<"initial obj func "<<val<<endl;
  
  float *pvn=new float[nnx*nnz](); read("pvn",pvn,nnx*nnz); 
  float *pvh=new float[nnx*nnz](); read("pvh",pvh,nnx*nnz); 
  float *pdel=new float[nnx*nnz](); read("pdel",pdel,nnx*nnz); 
 
  float *tvn=new float[nnx*nnz]();
  float *tvh=new float[nnx*nnz]();
  float *tdel=new float[nnx*nnz]();
 
  float s=(-1.)*s0;
  for(int i=1;i<10;++i){
   
   lin_comb(tvn,1.,vn,s,pvn,nnx*nnz);
   lin_comb(tvh,1.,vh,s,pvh,nnx*nnz);
   lin_comb(tdel,1.,del,s,pdel,nnx*nnz);
  
   val=objFunc5(d0,tvn,tvh,tdel,vn0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
   
   cout<<"Step "<<s<<" Obj func "<<val<<endl;
   s=a*s+b;
  }
  
  delete []vn;delete []vh;delete []del;
  delete []pvn;delete []pvh;delete []pdel;
  delete []tvn;delete []tvh;delete []tdel;
 }
 
 if(parameterize==6){
  float *v=new float[nnx*nnz]();
  float *vn=new float[nnx*nnz]();
  float *vh=new float[nnx*nnz]();
  
  if(padded==0){
   init_model("v",v,nx,nz,npad); 
   init_model("vn",vn,nx,nz,npad); 
   init_model("vh",vh,nx,nz,npad); 
  } 
  else{ 
   read("v",v,nnx*nnz);
   read("vn",vn,nnx*nnz);
   read("vh",vh,nnx*nnz);
  }
 
  scale(v,v,1e-3,nnx*nnz);
  scale(vn,vn,1e-3,nnx*nnz);
  scale(vh,vh,1e-3,nnx*nnz);

  float val=objFunc6(d0,v,vn,vh,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
  cout<<"initial obj func "<<val<<endl;
  
  float *pv=new float[nnx*nnz](); read("pv",pv,nnx*nnz); 
  float *pvn=new float[nnx*nnz](); read("pvn",pvn,nnx*nnz); 
  float *pvh=new float[nnx*nnz](); read("pvh",pvh,nnx*nnz); 
 
  float *tv=new float[nnx*nnz]();
  float *tvn=new float[nnx*nnz]();
  float *tvh=new float[nnx*nnz]();
 
  float s=(-1.)*s0;
  for(int i=1;i<10;++i){
   
   lin_comb(tv,1.,v,s,pv,nnx*nnz);
   lin_comb(tvn,1.,vn,s,pvn,nnx*nnz);
   lin_comb(tvh,1.,vh,s,pvh,nnx*nnz);
  
   val=objFunc6(d0,tv,tvn,tvh,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,m);
   
   cout<<"Step "<<s<<" Obj func "<<val<<endl;
   s=a*s+b;
  }
  
  delete []v;delete []vn;delete []vh;
  delete []pv;delete []pvn;delete []pvh;
  delete []tv;delete []tvn;delete []tvh;
 }
 
 delete []wavelet;delete []d0;delete []sloc;delete []rloc;delete []taper;
 delete []m;

 myio_close();
 return 0;
}
