//Following musa1 in sep150
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <iostream>
#include "mylib.h"
#include "wave.h"
#include "solvers.h"

float line_search0(float *c11,float *c13,float *c33,float &f,float *ggc11,float *ggc13,float *ggc33,float f0,const float *gc11,const float *gc13,const float *gc33,const float *pc11,const float *pc13,const float *pc33,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gc11,pc11,nnx*nnz)+dot_product(gc13,pc13,nnx*nnz)+dot_product(gc33,pc33,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew;

 float *tc11=new float[nnx*nnz]();
 float *tc13=new float[nnx*nnz]();
 float *tc33=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tc11,1.,c11,anew,pc11,nnx*nnz);
  lin_comb(tc13,1.,c13,anew,pc13,nnx*nnz);
  lin_comb(tc33,1.,c33,anew,pc33,nnx*nnz);
  
  fnew=objFuncGradient0(ggc11,ggc13,ggc33,d0,tc11,tc13,tc33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggc11,pc11,nnx*nnz)+dot_product(ggc13,pc13,nnx*nnz)+dot_product(ggc33,pc33,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom0(c11,c13,c33,f,ggc11,ggc13,ggc33,aold,anew,fold,fnew,gold,gnew,f0,g0,pc11,pc13,pc33,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tc11;delete []tc13;delete []tc33;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;
   memcpy(c11,tc11,nnx*nnz*sizeof(float));
   memcpy(c13,tc13,nnx*nnz*sizeof(float));
   memcpy(c33,tc33,nnx*nnz*sizeof(float));
   delete []tc11;delete []tc13;delete []tc33;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom0(c11,c13,c33,f,ggc11,ggc13,ggc33,anew,aold,fnew,fold,gnew,gold,f0,g0,pc11,pc13,pc33,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tc11;delete []tc13;delete []tc33;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;
 
 memcpy(c11,tc11,nnx*nnz*sizeof(float));
 memcpy(c13,tc13,nnx*nnz*sizeof(float));
 memcpy(c33,tc33,nnx*nnz*sizeof(float));

 delete []tc11;delete []tc13;delete []tc33;

 return a;
}

float zoom0(float *c11,float *c13,float *c33,float &f,float *ggc11,float *ggc13,float *ggc33,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pc11,const float *pc13,const float *pc33,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *c11tr=new float[nnx*nnz]();
 float *c13tr=new float[nnx*nnz]();
 float *c33tr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(c11tr,1.,c11,atr,pc11,nnx*nnz);
  lin_comb(c13tr,1.,c13,atr,pc13,nnx*nnz);
  lin_comb(c33tr,1.,c33,atr,pc33,nnx*nnz);
  
  ftr=objFuncGradient0(ggc11,ggc13,ggc33,d0,c11tr,c13tr,c33tr,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggc11,pc11,nnx*nnz)+dot_product(ggc13,pc13,nnx*nnz)+dot_product(ggc33,pc33,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(c11,c11tr,nnx*nnz*sizeof(float));
    memcpy(c13,c13tr,nnx*nnz*sizeof(float));
    memcpy(c33,c33tr,nnx*nnz*sizeof(float));
    
    delete []c11tr;delete []c13tr;delete []c33tr;

    f=ftr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(c11,c11tr,nnx*nnz*sizeof(float));
 memcpy(c13,c13tr,nnx*nnz*sizeof(float));
 memcpy(c33,c33tr,nnx*nnz*sizeof(float));
 
 delete []c11tr;delete []c13tr;delete []c33tr;

 f=ftr;
 return atr;
}

float line_search01(float *c11,float *c13,float *c33,float &f,float &datafit,float *ggc11,float *ggc13,float *ggc33,float f0,const float *gc11,const float *gc13,const float *gc33,const float *pc11,const float *pc13,const float *pc33,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gc11,pc11,nnx*nnz)+dot_product(gc13,pc13,nnx*nnz)+dot_product(gc33,pc33,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew,dfitnew,reg;

 float *tc11=new float[nnx*nnz]();
 float *tc13=new float[nnx*nnz]();
 float *tc33=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tc11,1.,c11,anew,pc11,nnx*nnz);
  lin_comb(tc13,1.,c13,anew,pc13,nnx*nnz);
  lin_comb(tc33,1.,c33,anew,pc33,nnx*nnz);
  
  fnew=objFuncGradient01(dfitnew,reg,ggc11,ggc13,ggc33,d0,tc11,tc13,tc33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
  
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggc11,pc11,nnx*nnz)+dot_product(ggc13,pc13,nnx*nnz)+dot_product(ggc33,pc33,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom01(c11,c13,c33,f,datafit,ggc11,ggc13,ggc33,aold,anew,fold,fnew,gold,gnew,f0,g0,pc11,pc13,pc33,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tc11;delete []tc13;delete []tc33;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;datafit=dfitnew;
   memcpy(c11,tc11,nnx*nnz*sizeof(float));
   memcpy(c13,tc13,nnx*nnz*sizeof(float));
   memcpy(c33,tc33,nnx*nnz*sizeof(float));
   delete []tc11;delete []tc13;delete []tc33;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom01(c11,c13,c33,f,datafit,ggc11,ggc13,ggc33,anew,aold,fnew,fold,gnew,gold,f0,g0,pc11,pc13,pc33,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tc11;delete []tc13;delete []tc33;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;datafit=dfitnew;
 
 memcpy(c11,tc11,nnx*nnz*sizeof(float));
 memcpy(c13,tc13,nnx*nnz*sizeof(float));
 memcpy(c33,tc33,nnx*nnz*sizeof(float));

 delete []tc11;delete []tc13;delete []tc33;

 return a;
}

float zoom01(float *c11,float *c13,float *c33,float &f,float &datafit,float *ggc11,float *ggc13,float *ggc33,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pc11,const float *pc13,const float *pc33,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *c11tr=new float[nnx*nnz]();
 float *c13tr=new float[nnx*nnz]();
 float *c33tr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr,dfittr,reg;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(c11tr,1.,c11,atr,pc11,nnx*nnz);
  lin_comb(c13tr,1.,c13,atr,pc13,nnx*nnz);
  lin_comb(c33tr,1.,c33,atr,pc33,nnx*nnz);
  
  ftr=objFuncGradient01(dfittr,reg,ggc11,ggc13,ggc33,d0,c11tr,c13tr,c33tr,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggc11,pc11,nnx*nnz)+dot_product(ggc13,pc13,nnx*nnz)+dot_product(ggc33,pc33,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(c11,c11tr,nnx*nnz*sizeof(float));
    memcpy(c13,c13tr,nnx*nnz*sizeof(float));
    memcpy(c33,c33tr,nnx*nnz*sizeof(float));
    
    delete []c11tr;delete []c13tr;delete []c33tr;

    f=ftr;datafit=dfittr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(c11,c11tr,nnx*nnz*sizeof(float));
 memcpy(c13,c13tr,nnx*nnz*sizeof(float));
 memcpy(c33,c33tr,nnx*nnz*sizeof(float));
 
 delete []c11tr;delete []c13tr;delete []c33tr;

 f=ftr;datafit=dfittr;
 return atr;
}

float line_search1(float *v,float *eps,float *del,float &f,float *ggv,float *ggeps,float *ggdel,float f0,const float *gv,const float *geps,const float *gdel,const float *pv,const float *peps,const float *pdel,float v0,float eps0,float del0,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gv,pv,nnx*nnz)+dot_product(geps,peps,nnx*nnz)+dot_product(gdel,pdel,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew;

 float *tv=new float[nnx*nnz]();
 float *teps=new float[nnx*nnz]();
 float *tdel=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tv,1.,v,anew,pv,nnx*nnz);
  lin_comb(teps,1.,eps,anew,peps,nnx*nnz);
  lin_comb(tdel,1.,del,anew,pdel,nnx*nnz);
  
  fnew=objFuncGradient1(ggv,ggeps,ggdel,d0,tv,teps,tdel,v0,eps0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggv,pv,nnx*nnz)+dot_product(ggeps,peps,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom1(v,eps,del,f,ggv,ggeps,ggdel,aold,anew,fold,fnew,gold,gnew,f0,g0,pv,peps,pdel,v0,eps0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tv;delete []teps;delete []tdel;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;
   memcpy(v,tv,nnx*nnz*sizeof(float));
   memcpy(eps,teps,nnx*nnz*sizeof(float));
   memcpy(del,tdel,nnx*nnz*sizeof(float));
   delete []tv;delete []teps;delete []tdel;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom1(v,eps,del,f,ggv,ggeps,ggdel,anew,aold,fnew,fold,gnew,gold,f0,g0,pv,peps,pdel,v0,eps0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tv;delete []teps;delete []tdel;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;
 
 memcpy(v,tv,nnx*nnz*sizeof(float));
 memcpy(eps,teps,nnx*nnz*sizeof(float));
 memcpy(del,tdel,nnx*nnz*sizeof(float));

 delete []tv;delete []teps;delete []tdel;

 return a;
}

float zoom1(float *v,float *eps,float *del,float &f,float *ggv,float *ggeps,float *ggdel,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pv,const float *peps,const float *pdel,float v0,float eps0,float del0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vtr=new float[nnx*nnz]();
 float *epstr=new float[nnx*nnz]();
 float *deltr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vtr,1.,v,atr,pv,nnx*nnz);
  lin_comb(epstr,1.,eps,atr,peps,nnx*nnz);
  lin_comb(deltr,1.,del,atr,pdel,nnx*nnz);

  ftr=objFuncGradient1(ggv,ggeps,ggdel,d0,vtr,epstr,deltr,v0,eps0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggv,pv,nnx*nnz)+dot_product(ggeps,peps,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(v,vtr,nnx*nnz*sizeof(float));
    memcpy(eps,epstr,nnx*nnz*sizeof(float));
    memcpy(del,deltr,nnx*nnz*sizeof(float));
    
    delete []vtr;delete []epstr;delete []deltr;

    f=ftr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(v,vtr,nnx*nnz*sizeof(float));
 memcpy(eps,epstr,nnx*nnz*sizeof(float));
 memcpy(del,deltr,nnx*nnz*sizeof(float));
 
 delete []vtr;delete []epstr;delete []deltr;

 f=ftr;
 return atr;
}

float line_search11(float *v,float *eps,float *del,float &f,float &datafit,float *ggv,float *ggeps,float *ggdel,float f0,const float *gv,const float *geps,const float *gdel,const float *pv,const float *peps,const float *pdel,float v0,float eps0,float del0,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gv,pv,nnx*nnz)+dot_product(geps,peps,nnx*nnz)+dot_product(gdel,pdel,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew,dfitnew,reg;

 float *tv=new float[nnx*nnz]();
 float *teps=new float[nnx*nnz]();
 float *tdel=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tv,1.,v,anew,pv,nnx*nnz);
  lin_comb(teps,1.,eps,anew,peps,nnx*nnz);
  lin_comb(tdel,1.,del,anew,pdel,nnx*nnz);
  
  fnew=objFuncGradient11(dfitnew,reg,ggv,ggeps,ggdel,d0,tv,teps,tdel,v0,eps0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggv,pv,nnx*nnz)+dot_product(ggeps,peps,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom11(v,eps,del,f,datafit,ggv,ggeps,ggdel,aold,anew,fold,fnew,gold,gnew,f0,g0,pv,peps,pdel,v0,eps0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tv;delete []teps;delete []tdel;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;datafit=dfitnew;
   memcpy(v,tv,nnx*nnz*sizeof(float));
   memcpy(eps,teps,nnx*nnz*sizeof(float));
   memcpy(del,tdel,nnx*nnz*sizeof(float));
   delete []tv;delete []teps;delete []tdel;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom11(v,eps,del,f,datafit,ggv,ggeps,ggdel,anew,aold,fnew,fold,gnew,gold,f0,g0,pv,peps,pdel,v0,eps0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tv;delete []teps;delete []tdel;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;datafit=dfitnew;
 
 memcpy(v,tv,nnx*nnz*sizeof(float));
 memcpy(eps,teps,nnx*nnz*sizeof(float));
 memcpy(del,tdel,nnx*nnz*sizeof(float));

 delete []tv;delete []teps;delete []tdel;

 return a;
}

float zoom11(float *v,float *eps,float *del,float &f,float &datafit,float *ggv,float *ggeps,float *ggdel,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pv,const float *peps,const float *pdel,float v0,float eps0,float del0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vtr=new float[nnx*nnz]();
 float *epstr=new float[nnx*nnz]();
 float *deltr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr,dfittr,reg;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vtr,1.,v,atr,pv,nnx*nnz);
  lin_comb(epstr,1.,eps,atr,peps,nnx*nnz);
  lin_comb(deltr,1.,del,atr,pdel,nnx*nnz);

  ftr=objFuncGradient11(dfittr,reg,ggv,ggeps,ggdel,d0,vtr,epstr,deltr,v0,eps0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggv,pv,nnx*nnz)+dot_product(ggeps,peps,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(v,vtr,nnx*nnz*sizeof(float));
    memcpy(eps,epstr,nnx*nnz*sizeof(float));
    memcpy(del,deltr,nnx*nnz*sizeof(float));
    
    delete []vtr;delete []epstr;delete []deltr;

    f=ftr;datafit=dfittr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(v,vtr,nnx*nnz*sizeof(float));
 memcpy(eps,epstr,nnx*nnz*sizeof(float));
 memcpy(del,deltr,nnx*nnz*sizeof(float));
 
 delete []vtr;delete []epstr;delete []deltr;

 f=ftr;datafit=dfittr;
 return atr;
}

float line_search2(float *vn,float *eta,float *del,float &f,float *ggvn,float *ggeta,float *ggdel,float f0,const float *gvn,const float *geta,const float *gdel,const float *pvn,const float *peta,const float *pdel,float vn0,float eta0,float del0,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gvn,pvn,nnx*nnz)+dot_product(geta,peta,nnx*nnz)+dot_product(gdel,pdel,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew;

 float *tvn=new float[nnx*nnz]();
 float *teta=new float[nnx*nnz]();
 float *tdel=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tvn,1.,vn,anew,pvn,nnx*nnz);
  lin_comb(teta,1.,eta,anew,peta,nnx*nnz);
  lin_comb(tdel,1.,del,anew,pdel,nnx*nnz);
  
  fnew=objFuncGradient2(ggvn,ggeta,ggdel,d0,tvn,teta,tdel,vn0,eta0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggeta,peta,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom2(vn,eta,del,f,ggvn,ggeta,ggdel,aold,anew,fold,fnew,gold,gnew,f0,g0,pvn,peta,pdel,vn0,eta0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tvn;delete []teta;delete []tdel;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;
   memcpy(vn,tvn,nnx*nnz*sizeof(float));
   memcpy(eta,teta,nnx*nnz*sizeof(float));
   memcpy(del,tdel,nnx*nnz*sizeof(float));
   delete []tvn;delete []teta;delete []tdel;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom2(vn,eta,del,f,ggvn,ggeta,ggdel,anew,aold,fnew,fold,gnew,gold,f0,g0,pvn,peta,pdel,vn0,eta0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tvn;delete []teta;delete []tdel;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;
 
 memcpy(vn,tvn,nnx*nnz*sizeof(float));
 memcpy(eta,teta,nnx*nnz*sizeof(float));
 memcpy(del,tdel,nnx*nnz*sizeof(float));

 delete []tvn;delete []teta;delete []tdel;

 return a;
}

float zoom2(float *vn,float *eta,float *del,float &f,float *ggvn,float *ggeta,float *ggdel,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pvn,const float *peta,const float *pdel,float vn0,float eta0,float del0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vntr=new float[nnx*nnz]();
 float *etatr=new float[nnx*nnz]();
 float *deltr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vntr,1.,vn,atr,pvn,nnx*nnz);
  lin_comb(etatr,1.,eta,atr,peta,nnx*nnz);
  lin_comb(deltr,1.,del,atr,pdel,nnx*nnz);

  ftr=objFuncGradient2(ggvn,ggeta,ggdel,d0,vntr,etatr,deltr,vn0,eta0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggeta,peta,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steta of zoom\n",i);
    memcpy(vn,vntr,nnx*nnz*sizeof(float));
    memcpy(eta,etatr,nnx*nnz*sizeof(float));
    memcpy(del,deltr,nnx*nnz*sizeof(float));
    
    delete []vntr;delete []etatr;delete []deltr;

    f=ftr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(vn,vntr,nnx*nnz*sizeof(float));
 memcpy(eta,etatr,nnx*nnz*sizeof(float));
 memcpy(del,deltr,nnx*nnz*sizeof(float));
 
 delete []vntr;delete []etatr;delete []deltr;

 f=ftr;
 return atr;
}

float line_search21(float *vn,float *eta,float *del,float &f,float &datafit,float *ggvn,float *ggeta,float *ggdel,float f0,const float *gvn,const float *geta,const float *gdel,const float *pvn,const float *peta,const float *pdel,float vn0,float eta0,float del0,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gvn,pvn,nnx*nnz)+dot_product(geta,peta,nnx*nnz)+dot_product(gdel,pdel,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew,dfitnew,reg;

 float *tvn=new float[nnx*nnz]();
 float *teta=new float[nnx*nnz]();
 float *tdel=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tvn,1.,vn,anew,pvn,nnx*nnz);
  lin_comb(teta,1.,eta,anew,peta,nnx*nnz);
  lin_comb(tdel,1.,del,anew,pdel,nnx*nnz);
  
  fnew=objFuncGradient21(dfitnew,reg,ggvn,ggeta,ggdel,d0,tvn,teta,tdel,vn0,eta0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggeta,peta,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom21(vn,eta,del,f,datafit,ggvn,ggeta,ggdel,aold,anew,fold,fnew,gold,gnew,f0,g0,pvn,peta,pdel,vn0,eta0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tvn;delete []teta;delete []tdel;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;datafit=dfitnew;
   memcpy(vn,tvn,nnx*nnz*sizeof(float));
   memcpy(eta,teta,nnx*nnz*sizeof(float));
   memcpy(del,tdel,nnx*nnz*sizeof(float));
   delete []tvn;delete []teta;delete []tdel;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom21(vn,eta,del,f,datafit,ggvn,ggeta,ggdel,anew,aold,fnew,fold,gnew,gold,f0,g0,pvn,peta,pdel,vn0,eta0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tvn;delete []teta;delete []tdel;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;datafit=dfitnew;
 
 memcpy(vn,tvn,nnx*nnz*sizeof(float));
 memcpy(eta,teta,nnx*nnz*sizeof(float));
 memcpy(del,tdel,nnx*nnz*sizeof(float));

 delete []tvn;delete []teta;delete []tdel;

 return a;
}

float zoom21(float *vn,float *eta,float *del,float &f,float &datafit,float *ggvn,float *ggeta,float *ggdel,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pvn,const float *peta,const float *pdel,float vn0,float eta0,float del0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vntr=new float[nnx*nnz]();
 float *etatr=new float[nnx*nnz]();
 float *deltr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr,dfittr,reg;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vntr,1.,vn,atr,pvn,nnx*nnz);
  lin_comb(etatr,1.,eta,atr,peta,nnx*nnz);
  lin_comb(deltr,1.,del,atr,pdel,nnx*nnz);

  ftr=objFuncGradient21(dfittr,reg,ggvn,ggeta,ggdel,d0,vntr,etatr,deltr,vn0,eta0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggeta,peta,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(vn,vntr,nnx*nnz*sizeof(float));
    memcpy(eta,etatr,nnx*nnz*sizeof(float));
    memcpy(del,deltr,nnx*nnz*sizeof(float));
    
    delete []vntr;delete []etatr;delete []deltr;

    f=ftr;datafit=dfittr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(vn,vntr,nnx*nnz*sizeof(float));
 memcpy(eta,etatr,nnx*nnz*sizeof(float));
 memcpy(del,deltr,nnx*nnz*sizeof(float));
 
 delete []vntr;delete []etatr;delete []deltr;

 f=ftr;datafit=dfittr;
 return atr;
}

float line_search3(float *vh,float *eps,float *eta,float &f,float *ggvh,float *ggeps,float *ggeta,float f0,const float *gvh,const float *geps,const float *geta,const float *pvh,const float *peps,const float *peta,float vh0,float eps0,float eta0,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gvh,pvh,nnx*nnz)+dot_product(geps,peps,nnx*nnz)+dot_product(geta,peta,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew;

 float *tvh=new float[nnx*nnz]();
 float *teps=new float[nnx*nnz]();
 float *teta=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tvh,1.,vh,anew,pvh,nnx*nnz);
  lin_comb(teps,1.,eps,anew,peps,nnx*nnz);
  lin_comb(teta,1.,eta,anew,peta,nnx*nnz);
  
  fnew=objFuncGradient3(ggvh,ggeps,ggeta,d0,tvh,teps,teta,vh0,eps0,eta0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggeps,peps,nnx*nnz)+dot_product(ggeta,peta,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom3(vh,eps,eta,f,ggvh,ggeps,ggeta,aold,anew,fold,fnew,gold,gnew,f0,g0,pvh,peps,peta,vh0,eps0,eta0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tvh;delete []teps;delete []teta;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;
   memcpy(vh,tvh,nnx*nnz*sizeof(float));
   memcpy(eps,teps,nnx*nnz*sizeof(float));
   memcpy(eta,teta,nnx*nnz*sizeof(float));
   delete []tvh;delete []teps;delete []teta;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom3(vh,eps,eta,f,ggvh,ggeps,ggeta,anew,aold,fnew,fold,gnew,gold,f0,g0,pvh,peps,peta,vh0,eps0,eta0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tvh;delete []teps;delete []teta;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;
 
 memcpy(vh,tvh,nnx*nnz*sizeof(float));
 memcpy(eps,teps,nnx*nnz*sizeof(float));
 memcpy(eta,teta,nnx*nnz*sizeof(float));

 delete []tvh;delete []teps;delete []teta;

 return a;
}

float zoom3(float *vh,float *eps,float *eta,float &f,float *ggvh,float *ggeps,float *ggeta,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pvh,const float *peps,const float *peta,float vh0,float eps0,float eta0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vhtr=new float[nnx*nnz]();
 float *epstr=new float[nnx*nnz]();
 float *etatr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vhtr,1.,vh,atr,pvh,nnx*nnz);
  lin_comb(epstr,1.,eps,atr,peps,nnx*nnz);
  lin_comb(etatr,1.,eta,atr,peta,nnx*nnz);

  ftr=objFuncGradient3(ggvh,ggeps,ggeta,d0,vhtr,epstr,etatr,vh0,eps0,eta0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggeps,peps,nnx*nnz)+dot_product(ggeta,peta,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(vh,vhtr,nnx*nnz*sizeof(float));
    memcpy(eps,epstr,nnx*nnz*sizeof(float));
    memcpy(eta,etatr,nnx*nnz*sizeof(float));
    
    delete []vhtr;delete []epstr;delete []etatr;

    f=ftr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(vh,vhtr,nnx*nnz*sizeof(float));
 memcpy(eps,epstr,nnx*nnz*sizeof(float));
 memcpy(eta,etatr,nnx*nnz*sizeof(float));
 
 delete []vhtr;delete []epstr;delete []etatr;

 f=ftr;
 return atr;
}

float line_search31(float *vh,float *eps,float *eta,float &f,float &datafit,float *ggvh,float *ggeps,float *ggeta,float f0,const float *gvh,const float *geps,const float *geta,const float *pvh,const float *peps,const float *peta,float vh0,float eps0,float eta0,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gvh,pvh,nnx*nnz)+dot_product(geps,peps,nnx*nnz)+dot_product(geta,peta,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew,dfitnew,reg;

 float *tvh=new float[nnx*nnz]();
 float *teps=new float[nnx*nnz]();
 float *teta=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tvh,1.,vh,anew,pvh,nnx*nnz);
  lin_comb(teps,1.,eps,anew,peps,nnx*nnz);
  lin_comb(teta,1.,eta,anew,peta,nnx*nnz);
  
  fnew=objFuncGradient31(dfitnew,reg,ggvh,ggeps,ggeta,d0,tvh,teps,teta,vh0,eps0,eta0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggeps,peps,nnx*nnz)+dot_product(ggeta,peta,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom31(vh,eps,eta,f,datafit,ggvh,ggeps,ggeta,aold,anew,fold,fnew,gold,gnew,f0,g0,pvh,peps,peta,vh0,eps0,eta0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tvh;delete []teps;delete []teta;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;datafit=dfitnew;
   memcpy(vh,tvh,nnx*nnz*sizeof(float));
   memcpy(eps,teps,nnx*nnz*sizeof(float));
   memcpy(eta,teta,nnx*nnz*sizeof(float));
   delete []tvh;delete []teps;delete []teta;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom31(vh,eps,eta,f,datafit,ggvh,ggeps,ggeta,anew,aold,fnew,fold,gnew,gold,f0,g0,pvh,peps,peta,vh0,eps0,eta0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tvh;delete []teps;delete []teta;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;datafit=dfitnew;
 
 memcpy(vh,tvh,nnx*nnz*sizeof(float));
 memcpy(eps,teps,nnx*nnz*sizeof(float));
 memcpy(eta,teta,nnx*nnz*sizeof(float));

 delete []tvh;delete []teps;delete []teta;

 return a;
}

float zoom31(float *vh,float *eps,float *eta,float &f,float &datafit,float *ggvh,float *ggeps,float *ggeta,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pvh,const float *peps,const float *peta,float vh0,float eps0,float eta0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vhtr=new float[nnx*nnz]();
 float *epstr=new float[nnx*nnz]();
 float *etatr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr,dfittr,reg;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vhtr,1.,vh,atr,pvh,nnx*nnz);
  lin_comb(epstr,1.,eps,atr,peps,nnx*nnz);
  lin_comb(etatr,1.,eta,atr,peta,nnx*nnz);

  ftr=objFuncGradient31(dfittr,reg,ggvh,ggeps,ggeta,d0,vhtr,epstr,etatr,vh0,eps0,eta0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggeps,peps,nnx*nnz)+dot_product(ggeta,peta,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(vh,vhtr,nnx*nnz*sizeof(float));
    memcpy(eps,epstr,nnx*nnz*sizeof(float));
    memcpy(eta,etatr,nnx*nnz*sizeof(float));
    
    delete []vhtr;delete []epstr;delete []etatr;

    f=ftr;datafit=dfittr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(vh,vhtr,nnx*nnz*sizeof(float));
 memcpy(eps,epstr,nnx*nnz*sizeof(float));
 memcpy(eta,etatr,nnx*nnz*sizeof(float));
 
 delete []vhtr;delete []epstr;delete []etatr;

 f=ftr;datafit=dfittr;
 return atr;
}

float line_search4(float *v,float *vh,float *del,float &f,float *ggv,float *ggvh,float *ggdel,float f0,const float *gv,const float *gvh,const float *gdel,const float *pv,const float *pvh,const float *pdel,float v0,float vh0,float del0,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gv,pv,nnx*nnz)+dot_product(gvh,pvh,nnx*nnz)+dot_product(gdel,pdel,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew;

 float *tv=new float[nnx*nnz]();
 float *tvh=new float[nnx*nnz]();
 float *tdel=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tv,1.,v,anew,pv,nnx*nnz);
  lin_comb(tvh,1.,vh,anew,pvh,nnx*nnz);
  lin_comb(tdel,1.,del,anew,pdel,nnx*nnz);
  
  fnew=objFuncGradient4(ggv,ggvh,ggdel,d0,tv,tvh,tdel,v0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggv,pv,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom4(v,vh,del,f,ggv,ggvh,ggdel,aold,anew,fold,fnew,gold,gnew,f0,g0,pv,pvh,pdel,v0,vh0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tv;delete []tvh;delete []tdel;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;
   memcpy(v,tv,nnx*nnz*sizeof(float));
   memcpy(vh,tvh,nnx*nnz*sizeof(float));
   memcpy(del,tdel,nnx*nnz*sizeof(float));
   delete []tv;delete []tvh;delete []tdel;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom4(v,vh,del,f,ggv,ggvh,ggdel,anew,aold,fnew,fold,gnew,gold,f0,g0,pv,pvh,pdel,v0,vh0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tv;delete []tvh;delete []tdel;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;
 
 memcpy(v,tv,nnx*nnz*sizeof(float));
 memcpy(vh,tvh,nnx*nnz*sizeof(float));
 memcpy(del,tdel,nnx*nnz*sizeof(float));

 delete []tv;delete []tvh;delete []tdel;

 return a;
}

float zoom4(float *v,float *vh,float *del,float &f,float *ggv,float *ggvh,float *ggdel,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pv,const float *pvh,const float *pdel,float v0,float vh0,float del0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vtr=new float[nnx*nnz]();
 float *vhtr=new float[nnx*nnz]();
 float *deltr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vtr,1.,v,atr,pv,nnx*nnz);
  lin_comb(vhtr,1.,vh,atr,pvh,nnx*nnz);
  lin_comb(deltr,1.,del,atr,pdel,nnx*nnz);

  ftr=objFuncGradient4(ggv,ggvh,ggdel,d0,vtr,vhtr,deltr,v0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggv,pv,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d stvh of zoom\n",i);
    memcpy(v,vtr,nnx*nnz*sizeof(float));
    memcpy(vh,vhtr,nnx*nnz*sizeof(float));
    memcpy(del,deltr,nnx*nnz*sizeof(float));
    
    delete []vtr;delete []vhtr;delete []deltr;

    f=ftr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(v,vtr,nnx*nnz*sizeof(float));
 memcpy(vh,vhtr,nnx*nnz*sizeof(float));
 memcpy(del,deltr,nnx*nnz*sizeof(float));
 
 delete []vtr;delete []vhtr;delete []deltr;

 f=ftr;
 return atr;
}

float line_search41(float *v,float *vh,float *del,float &f,float &datafit,float *ggv,float *ggvh,float *ggdel,float f0,const float *gv,const float *gvh,const float *gdel,const float *pv,const float *pvh,const float *pdel,float v0,float vh0,float del0,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gv,pv,nnx*nnz)+dot_product(gvh,pvh,nnx*nnz)+dot_product(gdel,pdel,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew,dfitnew,reg;

 float *tv=new float[nnx*nnz]();
 float *tvh=new float[nnx*nnz]();
 float *tdel=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tv,1.,v,anew,pv,nnx*nnz);
  lin_comb(tvh,1.,vh,anew,pvh,nnx*nnz);
  lin_comb(tdel,1.,del,anew,pdel,nnx*nnz);
  
  fnew=objFuncGradient41(dfitnew,reg,ggv,ggvh,ggdel,d0,tv,tvh,tdel,v0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggv,pv,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom41(v,vh,del,f,datafit,ggv,ggvh,ggdel,aold,anew,fold,fnew,gold,gnew,f0,g0,pv,pvh,pdel,v0,vh0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tv;delete []tvh;delete []tdel;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;datafit=dfitnew;
   memcpy(v,tv,nnx*nnz*sizeof(float));
   memcpy(vh,tvh,nnx*nnz*sizeof(float));
   memcpy(del,tdel,nnx*nnz*sizeof(float));
   delete []tv;delete []tvh;delete []tdel;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom41(v,vh,del,f,datafit,ggv,ggvh,ggdel,anew,aold,fnew,fold,gnew,gold,f0,g0,pv,pvh,pdel,v0,vh0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tv;delete []tvh;delete []tdel;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;datafit=dfitnew;
 
 memcpy(v,tv,nnx*nnz*sizeof(float));
 memcpy(vh,tvh,nnx*nnz*sizeof(float));
 memcpy(del,tdel,nnx*nnz*sizeof(float));

 delete []tv;delete []tvh;delete []tdel;

 return a;
}

float zoom41(float *v,float *vh,float *del,float &f,float &datafit,float *ggv,float *ggvh,float *ggdel,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pv,const float *pvh,const float *pdel,float v0,float vh0,float del0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vtr=new float[nnx*nnz]();
 float *vhtr=new float[nnx*nnz]();
 float *deltr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr,dfittr,reg;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vtr,1.,v,atr,pv,nnx*nnz);
  lin_comb(vhtr,1.,vh,atr,pvh,nnx*nnz);
  lin_comb(deltr,1.,del,atr,pdel,nnx*nnz);

  ftr=objFuncGradient41(dfittr,reg,ggv,ggvh,ggdel,d0,vtr,vhtr,deltr,v0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggv,pv,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(v,vtr,nnx*nnz*sizeof(float));
    memcpy(vh,vhtr,nnx*nnz*sizeof(float));
    memcpy(del,deltr,nnx*nnz*sizeof(float));
    
    delete []vtr;delete []vhtr;delete []deltr;

    f=ftr;datafit=dfittr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(v,vtr,nnx*nnz*sizeof(float));
 memcpy(vh,vhtr,nnx*nnz*sizeof(float));
 memcpy(del,deltr,nnx*nnz*sizeof(float));
 
 delete []vtr;delete []vhtr;delete []deltr;

 f=ftr;datafit=dfittr;
 return atr;
}

float line_search5(float *vn,float *vh,float *del,float &f,float *ggvn,float *ggvh,float *ggdel,float f0,const float *gvn,const float *gvh,const float *gdel,const float *pvn,const float *pvh,const float *pdel,float vn0,float vh0,float del0,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gvn,pvn,nnx*nnz)+dot_product(gvh,pvh,nnx*nnz)+dot_product(gdel,pdel,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew;

 float *tvn=new float[nnx*nnz]();
 float *tvh=new float[nnx*nnz]();
 float *tdel=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tvn,1.,vn,anew,pvn,nnx*nnz);
  lin_comb(tvh,1.,vh,anew,pvh,nnx*nnz);
  lin_comb(tdel,1.,del,anew,pdel,nnx*nnz);
  
  fnew=objFuncGradient5(ggvn,ggvh,ggdel,d0,tvn,tvh,tdel,vn0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom5(vn,vh,del,f,ggvn,ggvh,ggdel,aold,anew,fold,fnew,gold,gnew,f0,g0,pvn,pvh,pdel,vn0,vh0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tvn;delete []tvh;delete []tdel;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;
   memcpy(vn,tvn,nnx*nnz*sizeof(float));
   memcpy(vh,tvh,nnx*nnz*sizeof(float));
   memcpy(del,tdel,nnx*nnz*sizeof(float));
   delete []tvn;delete []tvh;delete []tdel;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom5(vn,vh,del,f,ggvn,ggvh,ggdel,anew,aold,fnew,fold,gnew,gold,f0,g0,pvn,pvh,pdel,vn0,vh0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tvn;delete []tvh;delete []tdel;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;
 
 memcpy(vn,tvn,nnx*nnz*sizeof(float));
 memcpy(vh,tvh,nnx*nnz*sizeof(float));
 memcpy(del,tdel,nnx*nnz*sizeof(float));

 delete []tvn;delete []tvh;delete []tdel;

 return a;
}

float zoom5(float *vn,float *vh,float *del,float &f,float *ggvn,float *ggvh,float *ggdel,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pvn,const float *pvh,const float *pdel,float vn0,float vh0,float del0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vntr=new float[nnx*nnz]();
 float *vhtr=new float[nnx*nnz]();
 float *deltr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vntr,1.,vn,atr,pvn,nnx*nnz);
  lin_comb(vhtr,1.,vh,atr,pvh,nnx*nnz);
  lin_comb(deltr,1.,del,atr,pdel,nnx*nnz);

  ftr=objFuncGradient5(ggvn,ggvh,ggdel,d0,vntr,vhtr,deltr,vn0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d stvh of zoom\n",i);
    memcpy(vn,vntr,nnx*nnz*sizeof(float));
    memcpy(vh,vhtr,nnx*nnz*sizeof(float));
    memcpy(del,deltr,nnx*nnz*sizeof(float));
    
    delete []vntr;delete []vhtr;delete []deltr;

    f=ftr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(vn,vntr,nnx*nnz*sizeof(float));
 memcpy(vh,vhtr,nnx*nnz*sizeof(float));
 memcpy(del,deltr,nnx*nnz*sizeof(float));
 
 delete []vntr;delete []vhtr;delete []deltr;

 f=ftr;
 return atr;
}

float line_search51(float *vn,float *vh,float *del,float &f,float &datafit,float *ggvn,float *ggvh,float *ggdel,float f0,const float *gvn,const float *gvh,const float *gdel,const float *pvn,const float *pvh,const float *pdel,float vn0,float vh0,float del0,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gvn,pvn,nnx*nnz)+dot_product(gvh,pvh,nnx*nnz)+dot_product(gdel,pdel,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew,dfitnew,reg;

 float *tvn=new float[nnx*nnz]();
 float *tvh=new float[nnx*nnz]();
 float *tdel=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tvn,1.,vn,anew,pvn,nnx*nnz);
  lin_comb(tvh,1.,vh,anew,pvh,nnx*nnz);
  lin_comb(tdel,1.,del,anew,pdel,nnx*nnz);
  
  fnew=objFuncGradient51(dfitnew,reg,ggvn,ggvh,ggdel,d0,tvn,tvh,tdel,vn0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom51(vn,vh,del,f,datafit,ggvn,ggvh,ggdel,aold,anew,fold,fnew,gold,gnew,f0,g0,pvn,pvh,pdel,vn0,vh0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tvn;delete []tvh;delete []tdel;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;datafit=dfitnew;
   memcpy(vn,tvn,nnx*nnz*sizeof(float));
   memcpy(vh,tvh,nnx*nnz*sizeof(float));
   memcpy(del,tdel,nnx*nnz*sizeof(float));
   delete []tvn;delete []tvh;delete []tdel;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom51(vn,vh,del,f,datafit,ggvn,ggvh,ggdel,anew,aold,fnew,fold,gnew,gold,f0,g0,pvn,pvh,pdel,vn0,vh0,del0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tvn;delete []tvh;delete []tdel;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;datafit=dfitnew;
 
 memcpy(vn,tvn,nnx*nnz*sizeof(float));
 memcpy(vh,tvh,nnx*nnz*sizeof(float));
 memcpy(del,tdel,nnx*nnz*sizeof(float));

 delete []tvn;delete []tvh;delete []tdel;

 return a;
}

float zoom51(float *vn,float *vh,float *del,float &f,float &datafit,float *ggvn,float *ggvh,float *ggdel,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pvn,const float *pvh,const float *pdel,float vn0,float vh0,float del0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vntr=new float[nnx*nnz]();
 float *vhtr=new float[nnx*nnz]();
 float *deltr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr,dfittr,reg;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vntr,1.,vn,atr,pvn,nnx*nnz);
  lin_comb(vhtr,1.,vh,atr,pvh,nnx*nnz);
  lin_comb(deltr,1.,del,atr,pdel,nnx*nnz);

  ftr=objFuncGradient51(dfittr,reg,ggvn,ggvh,ggdel,d0,vntr,vhtr,deltr,vn0,vh0,del0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz)+dot_product(ggdel,pdel,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(vn,vntr,nnx*nnz*sizeof(float));
    memcpy(vh,vhtr,nnx*nnz*sizeof(float));
    memcpy(del,deltr,nnx*nnz*sizeof(float));
    
    delete []vntr;delete []vhtr;delete []deltr;

    f=ftr;datafit=dfittr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(vn,vntr,nnx*nnz*sizeof(float));
 memcpy(vh,vhtr,nnx*nnz*sizeof(float));
 memcpy(del,deltr,nnx*nnz*sizeof(float));
 
 delete []vntr;delete []vhtr;delete []deltr;

 f=ftr;datafit=dfittr;
 return atr;
}

float line_search6(float *v,float *vn,float *vh,float &f,float *ggv,float *ggvn,float *ggvh,float f0,const float *gv,const float *gvn,const float *gvh,const float *pv,const float *pvn,const float *pvh,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gv,pv,nnx*nnz)+dot_product(gvn,pvn,nnx*nnz)+dot_product(gvh,pvh,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew;

 float *tv=new float[nnx*nnz]();
 float *tvn=new float[nnx*nnz]();
 float *tvh=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tv,1.,v,anew,pv,nnx*nnz);
  lin_comb(tvn,1.,vn,anew,pvn,nnx*nnz);
  lin_comb(tvh,1.,vh,anew,pvh,nnx*nnz);
  
  fnew=objFuncGradient6(ggv,ggvn,ggvh,d0,tv,tvn,tvh,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggv,pv,nnx*nnz)+dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom6(v,vn,vh,f,ggv,ggvn,ggvh,aold,anew,fold,fnew,gold,gnew,f0,g0,pv,pvn,pvh,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tv;delete []tvn;delete []tvh;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;
   memcpy(v,tv,nnx*nnz*sizeof(float));
   memcpy(vn,tvn,nnx*nnz*sizeof(float));
   memcpy(vh,tvh,nnx*nnz*sizeof(float));
   delete []tv;delete []tvn;delete []tvh;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom6(v,vn,vh,f,ggv,ggvn,ggvh,anew,aold,fnew,fold,gnew,gold,f0,g0,pv,pvn,pvh,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
   delete []tv;delete []tvn;delete []tvh;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;
 
 memcpy(v,tv,nnx*nnz*sizeof(float));
 memcpy(vn,tvn,nnx*nnz*sizeof(float));
 memcpy(vh,tvh,nnx*nnz*sizeof(float));

 delete []tv;delete []tvn;delete []tvh;

 return a;
}

float zoom6(float *v,float *vn,float *vh,float &f,float *ggv,float *ggvn,float *ggvh,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pv,const float *pvn,const float *pvh,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vtr=new float[nnx*nnz]();
 float *vntr=new float[nnx*nnz]();
 float *vhtr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vtr,1.,v,atr,pv,nnx*nnz);
  lin_comb(vntr,1.,vn,atr,pvn,nnx*nnz);
  lin_comb(vhtr,1.,vh,atr,pvh,nnx*nnz);

  ftr=objFuncGradient6(ggv,ggvn,ggvh,d0,vtr,vntr,vhtr,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggv,pv,nnx*nnz)+dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d stvn of zoom\n",i);
    memcpy(v,vtr,nnx*nnz*sizeof(float));
    memcpy(vn,vntr,nnx*nnz*sizeof(float));
    memcpy(vh,vhtr,nnx*nnz*sizeof(float));
    
    delete []vtr;delete []vntr;delete []vhtr;

    f=ftr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(v,vtr,nnx*nnz*sizeof(float));
 memcpy(vn,vntr,nnx*nnz*sizeof(float));
 memcpy(vh,vhtr,nnx*nnz*sizeof(float));
 
 delete []vtr;delete []vntr;delete []vhtr;

 f=ftr;
 return atr;
}

float line_search61(float *v,float *vn,float *vh,float &f,float &datafit,float *ggv,float *ggvn,float *ggvh,float f0,const float *gv,const float *gvn,const float *gvh,const float *pv,const float *pvn,const float *pvh,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gv,pv,nnx*nnz)+dot_product(gvn,pvn,nnx*nnz)+dot_product(gvh,pvh,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew,dfitnew,reg;

 float *tv=new float[nnx*nnz]();
 float *tvn=new float[nnx*nnz]();
 float *tvh=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tv,1.,v,anew,pv,nnx*nnz);
  lin_comb(tvn,1.,vn,anew,pvn,nnx*nnz);
  lin_comb(tvh,1.,vh,anew,pvh,nnx*nnz);
  
  fnew=objFuncGradient61(dfitnew,reg,ggv,ggvn,ggvh,d0,tv,tvn,tvh,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
 
  if(isnan(fnew) || isinf(fnew)) {
   anew*=0.2;
   continue;
  }

  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggv,pv,nnx*nnz)+dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz);
  
  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom61(v,vn,vh,f,datafit,ggv,ggvn,ggvh,aold,anew,fold,fnew,gold,gnew,f0,g0,pv,pvn,pvh,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tv;delete []tvn;delete []tvh;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;datafit=dfitnew;
   memcpy(v,tv,nnx*nnz*sizeof(float));
   memcpy(vn,tvn,nnx*nnz*sizeof(float));
   memcpy(vh,tvh,nnx*nnz*sizeof(float));
   delete []tv;delete []tvn;delete []tvh;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom61(v,vn,vh,f,datafit,ggv,ggvn,ggvh,anew,aold,fnew,fold,gnew,gold,f0,g0,pv,pvn,pvh,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
   delete []tv;delete []tvn;delete []tvh;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;datafit=dfitnew;
 
 memcpy(v,tv,nnx*nnz*sizeof(float));
 memcpy(vn,tvn,nnx*nnz*sizeof(float));
 memcpy(vh,tvh,nnx*nnz*sizeof(float));

 delete []tv;delete []tvn;delete []tvh;

 return a;
}

float zoom61(float *v,float *vn,float *vh,float &f,float &datafit,float *ggv,float *ggvn,float *ggvh,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pv,const float *pvn,const float *pvh,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *vtr=new float[nnx*nnz]();
 float *vntr=new float[nnx*nnz]();
 float *vhtr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr,dfittr,reg;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(vtr,1.,v,atr,pv,nnx*nnz);
  lin_comb(vntr,1.,vn,atr,pvn,nnx*nnz);
  lin_comb(vhtr,1.,vh,atr,pvh,nnx*nnz);

  ftr=objFuncGradient61(dfittr,reg,ggv,ggvn,ggvh,d0,vtr,vntr,vhtr,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate,wbottom,m,alpha,aa,lag,na,patch);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggv,pv,nnx*nnz)+dot_product(ggvn,pvn,nnx*nnz)+dot_product(ggvh,pvh,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(v,vtr,nnx*nnz*sizeof(float));
    memcpy(vn,vntr,nnx*nnz*sizeof(float));
    memcpy(vh,vhtr,nnx*nnz*sizeof(float));
    
    delete []vtr;delete []vntr;delete []vhtr;

    f=ftr;datafit=dfittr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(v,vtr,nnx*nnz*sizeof(float));
 memcpy(vn,vntr,nnx*nnz*sizeof(float));
 memcpy(vh,vhtr,nnx*nnz*sizeof(float));
 
 delete []vtr;delete []vntr;delete []vhtr;

 f=ftr;datafit=dfittr;
 return atr;
}

