#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "myio.h"
#include "mylib.h"
#include "init.h"
#include "helix.h"
#include "steer.h"

using namespace std;

void L(float *d,float *reg,float *m,bool adj,float eps,int *loc,int nwells,float **aa,int **lag,int *na,int *patch,int nnx,int nnz){
 ns_helixcon(reg,m,nnx*nnz,aa,lag,na,patch,adj);
 if(adj){
  scale(m,m,eps,nnx*nnz);
  for(int iz=0;iz<nnz;++iz){
   for(int iw=0;iw<nwells;++iw){
    m[loc[iw]+iz*nnx]+=d[iw+iz*nwells]; 
   }
  }
 }
 else{
  scale(reg,reg,eps,nnx*nnz);
  for(int iz=0;iz<nnz;++iz){
   for(int iw=0;iw<nwells;++iw){
    d[iw+iz*nwells]=m[loc[iw]+iz*nnx];
   }
  }
 }
return; 
}

int main(int argc,char **argv){
 myio_init(argc,argv);
 
 int nx=1600,nz=400,npad=28;
 int nnx=nx+2*npad,nnz=nz+2*npad,n=nnx*nnz;
 float dx=12.5,dz=12.5;

 float *m=new float[n](); init_model("m",m,nx,nz,npad); 
 float *m0=new float[n](); 
 init_model("m0",m0,nx,nz,npad); 
// scale(m0,m0,1e-3,n);scale(m,m,1e-3,n);
 
 int nwells; get_param("nwells",nwells);
 int *loc=new int[nwells]();
 for(int iw=0;iw<nwells;++iw) {
  loc[iw]=rand()%nnx;
//  cout<<loc[iw]<<endl;
 }

 int nd=nwells*nnz;
 float *d=new float[nd]();
 for(int iz=0;iz<nnz;++iz){
  for(int iw=0;iw<nwells;++iw){
   d[iw+iz*nwells]=m[loc[iw]+iz*nnx];
  }
 }

 float diptop; get_param("diptop",diptop);
 int ntop=npad+diptop/dz;
 int nfilters=1+(nnz-ntop)*nnx;

 int *na=new int[nfilters]();
 na[0]=8;
 
 int *patch=new int[nnx*nnz]();
 set(patch,0,ntop*nnx);

 float **aa=new float*[nfilters]();
 aa[0]=new float[na[0]]();

 int **lag=new int*[nfilters]();
 lag[0]=new int[na[0]]();
 
 float temp=lapfac(aa[0],lag[0],na[0],0.001,nnx,20);
 
 float *dip=new float[(nnz-ntop)*nnx]();
 read("dip",dip,(nnz-ntop)*nnx);

 for(int i=1;i<nfilters;++i){
  na[i]=3;
  aa[i]=new float[na[i]]();
  lag[i]=new int[na[i]]();
  steer0(aa[i],lag[i],na[i],dip[i-1],nnx); 
  patch[i-1+ntop*nnx]=i;
 }
 
 float *r=new float[nd]();
 float *gd=new float[nd]();
 float *sd=new float[nd]();

 float *reg=new float[n]();
 float *rreg=new float[n]();
 float *greg=new float[n]();
 float *sreg=new float[n]();

 float *g=new float[n]();
 float *s=new float[n]();

 float eps; get_param("eps",eps);

 int niter; get_param("niter",niter);

 write("im",m0,n);
 to_header("im","n1",nnx,"o1",-dx*npad,"d1",dx);
 to_header("im","n2",nnz,"o2",-dz*npad,"d2",dz);
 to_header("im","n3",niter+1,"o3",0.,"d3",1.);
 
 for(int i=0;i<niter;++i){
  memset(r,0,sizeof(float)*nd);
  memset(rreg,0,sizeof(float)*n);
  L(r,rreg,m0,false,eps,loc,nwells,aa,lag,na,patch,nnx,nnz);
  
  subtract(r,r,d,nd);
  subtract(rreg,rreg,reg,n);
  
  cout<<dot_product(r,r,nd)+dot_product(rreg,rreg,n)<<" "<<dot_product(r,r,nd)<<" "<<dot_product(rreg,rreg,n)<<endl;
  L(r,rreg,g,true,eps,loc,nwells,aa,lag,na,patch,nnx,nnz);
  L(gd,greg,g,false,eps,loc,nwells,aa,lag,na,patch,nnx,nnz);
 
  float alpha,beta;

  if(i==0){
   alpha=-(dot_product(gd,r,nd)+dot_product(greg,rreg,n))/(dot_product(gd,gd,nd)+dot_product(greg,greg,n));
   beta=0.;
  }
  else{
   float gdg=dot_product(gd,gd,nd)+dot_product(greg,greg,n);
   float sds=dot_product(sd,sd,nd)+dot_product(sreg,sreg,n);
   float gds=dot_product(gd,sd,nd)+dot_product(greg,sreg,n);
   float determ=gdg*sds*max(1.-gds*gds/gdg/sds,1e-10);
   float gdr=-dot_product(gd,r,nd)-dot_product(greg,rreg,n);
   float sdr=-dot_product(sd,r,nd)-dot_product(sreg,rreg,n);
   alpha=(sds*gdr-gds*sdr)/determ;
   beta=(gdg*sdr-gds*gdr)/determ;
//   fprintf(stderr,"gdg %f sds %f gds %f gdr %f sdr %f determ %f\n",gdg,sds,gds,gdr,sdr,determ);  
  }
//  cout<<"iter "<<i<<" alpha "<<alpha<<" beta "<<beta<<endl;
  
  lin_comb(s,alpha,g,beta,s,n);
  lin_comb(sd,alpha,gd,beta,sd,nd);
  lin_comb(sreg,alpha,greg,beta,sreg,n);
  
  add(m0,m0,s,n);
  add(r,r,sd,nd);
  add(rreg,rreg,sreg,n);
  
  write("im",m0,n,std::ios_base::app);
 }
 
 myio_close();
 return 0;
}
