#include <cmath>
#include <omp.h>
#include <cstdio>

#include "myio.h"
#include "mylib.h"
#include "init.h"
#include "wave.h"

using namespace std;

int main(int argc,char **argv){
 myio_init(argc,argv);

 int nnx,nnz,nx,nz,nt,npad;
 float ox,oz,ot,dx,dz,dt;
 init2d(nnx,nnz,nx,nz,nt,dx,dz,dt,ox,oz,ot,npad);

 float f;
 get_param("f",f);
 float *wavelet=new float[nt]();
 ricker(wavelet,f,nt,dt);
 
 float rate;
 get_param("rate",rate);
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 float *d=new float[nnx*nnz*nnt]();
 
 float *taper=new float[nnx*nnz]();
 init_abc(taper,nx,nz,npad);
 
 float v,eps,del;
 get_param("v",v,"eps",eps,"del",del);
 
 float *c11=new float[nnx*nnz](); set(c11,v*v*(1.+2.*eps),nnx*nnz);
 float *c13=new float[nnx*nnz](); set(c13,v*v*sqrt(1.+2.*del),nnx*nnz);
 float *c33=new float[nnx*nnz](); set(c33,v*v,nnx*nnz);
 
 float sx; get_param("sx",sx);
 float sz; get_param("sz",sz);
 int slocxz=(sx/dx+npad+1)+(sz/dz+npad+1)*nnx;

 modeling_f(d,c11,c13,c33,wavelet,slocxz,taper,nx,nz,nt,npad,dx,dz,dt,rate);

 write("data",d,nnx*nnz*nnt);
 to_header("data","n1",nnx,"o1",-dx*npad,"d1",dx);
 to_header("data","n2",nnz,"o2",-dz*npad,"d2",dz);
 to_header("data","n3",nnt,"o3",0.,"d3",rate);

 delete []wavelet;
 delete []c11;
 delete []c13;
 delete []c33;
 delete []d;
 delete []taper;

 myio_close();
 return 0;
}
