#include <omp.h>
#include <cstdlib>
#include <cstdio>
#include "mylib.h"
#include "init.h"
#include "wave.h"
#include "kernels.h"
#include "helix.h"

void modeling_f(float *d,const float *c11,const float *c13,const float *c33,const float *wavelet,int slocxz,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 fprintf(stderr,"Modeling\n");

 int ratio=rate/dt+0.5;
 int nnx=nx+2*npad,nnz=nz+2*npad;
 float dx2=dx*dx,dz2=dz*dz,dt2=dt*dt;
 
 cudaSetDevice(0);
 init_coeffs();

 float *d_d; cudaMalloc(&d_d,nnx*nnz*sizeof(float));
 float *d_c11; cudaMalloc(&d_c11,nnx*nnz*sizeof(float));
 float *d_c13; cudaMalloc(&d_c13,nnx*nnz*sizeof(float));
 float *d_c33; cudaMalloc(&d_c33,nnx*nnz*sizeof(float));
 cudaMemcpy(d_c11,c11,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 cudaMemcpy(d_c13,c13,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 cudaMemcpy(d_c33,c33,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 
 float *d_taper; cudaMalloc(&d_taper,nnx*nnz*sizeof(float));
 cudaMemcpy(d_taper,taper,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);

 float *p0; cudaMalloc(&p0,nnx*nnz*sizeof(float)); 
 float *p1; cudaMalloc(&p1,nnx*nnz*sizeof(float)); 
 float *q0; cudaMalloc(&q0,nnx*nnz*sizeof(float)); 
 float *q1; cudaMalloc(&q1,nnx*nnz*sizeof(float)); 

 dim3 block(BLOCK_DIM_X,BLOCK_DIM_Y);
 dim3 grid((nnx-2*RADIUS)/BLOCK_DIM_X,(nnz-2*RADIUS)/BLOCK_DIM_Y);

 cudaMemset(p0,0,nnx*nnz*sizeof(float));
 cudaMemset(q0,0,nnx*nnz*sizeof(float));
 cudaMemset(p1,0,nnx*nnz*sizeof(float));
 cudaMemset(q1,0,nnx*nnz*sizeof(float));

 inject<<<1,1>>>(p1,q1,wavelet[0],slocxz,dt2);

 abc<<<grid,block>>>(p1,q1,d_taper);
 
 if(ratio==1){
  record<<<grid,block>>>(d_d,p1,q1);
  cudaMemcpy(d+nnx*nnz,d_d,nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
 }
 
 for(int it=2;it<nt;++it){
  //fprintf(stderr,"Time step it=%d\n",it);

  forwardCD<<<grid,block>>>(p0,q0,p1,q1,d_c11,d_c13,d_c33,dx2,dz2,dt2);

  inject<<<1,1>>>(p0,q0,wavelet[it-1],slocxz,dt2);

  abc<<<grid,block>>>(p0,q0,d_taper);
  abc<<<grid,block>>>(p1,q1,d_taper);
  
  if(it%ratio==0){
   record<<<grid,block>>>(d_d,p0,q0);
   cudaMemcpy(d+(it/ratio)*nnx*nnz,d_d,nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
  }

  float *pt=p0; 
  p0=p1;
  p1=pt;
  pt=q0;
  q0=q1;
  q1=pt;
 }
  
 cudaFree(d_c11);cudaFree(d_c13);cudaFree(d_c33);
 cudaFree(d_taper);
 cudaFree(p0);cudaFree(p1);cudaFree(q0);cudaFree(q1);
 
 return;
}
