#include <omp.h>
#include <cstdio>
#include "mylib.h"
#include "init.h"
#include "wave.h"
#include "check.h"
#include "conversions.h"
#include "kernels.h"
#include "helix.h"

float objFunc0(const float *d0,float *c11,float *c13,float *c33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkCij(c11,c13,c33,nnx*nnz,m);
 float *tc11=new float[nnx*nnz](); scale(tc11,c11,1e6,nnx*nnz);
 float *tc13=new float[nnx*nnz](); scale(tc13,c13,1e6,nnx*nnz);
 float *tc33=new float[nnx*nnz](); scale(tc33,c33,1e6,nnx*nnz);
 double val=objFunc_f(d0,tc11,tc13,tc33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 delete []tc11;delete []tc13;delete []tc33;
 return val;
}

float objFunc01(float &datafit,float &reg,const float *d0,float *c11,float *c13,float *c33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkCij(c11,c13,c33,nnx*nnz,m);
 float *tc11=new float[nnx*nnz](); scale(tc11,c11,1e6,nnx*nnz);
 float *tc13=new float[nnx*nnz](); scale(tc13,c13,1e6,nnx*nnz);
 float *tc33=new float[nnx*nnz](); scale(tc33,c33,1e6,nnx*nnz);
 datafit=objFunc_f(d0,tc11,tc13,tc33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 delete []tc11;delete []tc13;delete []tc33;
 float *t11=new float[nnx*nnz](); ns_helixcon(t11,c11,nnx*nnz,aa,lag,na,patch,false);
 float *t13=new float[nnx*nnz](); ns_helixcon(t13,c13,nnx*nnz,aa,lag,na,patch,false);
 float *t33=new float[nnx*nnz](); ns_helixcon(t33,c33,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(t11,t11,nnx*nnz)+dot_product(t13,t13,nnx*nnz)+dot_product(t33,t33,nnx*nnz)); 
 delete []t11;delete []t13;delete []t33;
 return datafit+reg;
}

float objFunc1(const float *d0,float *v,float *eps,float *del,float v0,float eps0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEpsDel(eps,del,eps0,del0,nnx*nnz,m);
// for(int iz=0;iz<nnz;++iz){
//  for(int ix=0;ix<nnx;++ix){
//   if(m[ix+iz*nnx]==1){
//    del[ix+iz*nnx]=(del[ix-1+(iz-1)*nnx]+del[ix+(iz-1)*nnx]+del[ix+1+(iz-1)*nnx]+
//                    del[ix-1+iz*nnx]+del[ix+iz*nnx]+del[ix+1+iz*nnx]+
//	            	del[ix-1+(iz+1)*nnx]+del[ix+(iz+1)*nnx]+del[ix+1+(iz+1)*nnx])/9.;
//    eps[ix+iz*nnx]=(eps[ix-1+(iz-1)*nnx]+eps[ix+(iz-1)*nnx]+eps[ix+1+(iz-1)*nnx]+
//                    eps[ix-1+iz*nnx]+eps[ix+iz*nnx]+eps[ix+1+iz*nnx]+
//	            	eps[ix-1+(iz+1)*nnx]+eps[ix+(iz+1)*nnx]+eps[ix+1+(iz+1)*nnx])/9.;
//   }					
//  }
// }
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VEpsDel2Cij(c11,c13,c33,v,eps,del,v0,eps0,del0,nnx*nnz);
 double val=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 delete []c11;delete []c13;delete []c33;
 return val; 
}

float objFunc11(float &datafit,float &reg,const float *d0,float *v,float *eps,float *del,float v0,float eps0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEpsDel(eps,del,eps0,del0,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VEpsDel2Cij(c11,c13,c33,v,eps,del,v0,eps0,del0,nnx*nnz);
 datafit=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 float *tv=new float[nnx*nnz](); ns_helixcon(tv,v,nnx*nnz,aa,lag,na,patch,false);
 float *teps=new float[nnx*nnz](); ns_helixcon(teps,eps,nnx*nnz,aa,lag,na,patch,false);
 float *tdel=new float[nnx*nnz](); ns_helixcon(tdel,del,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tv,tv,nnx*nnz)+dot_product(teps,teps,nnx*nnz)+dot_product(tdel,tdel,nnx*nnz)); 
 delete []c11;delete []c13;delete []c33;
 delete []tv;delete []teps;delete []tdel;
 return datafit+reg; 
}

float objFunc2(const float *d0,float *vn,float *eta,float *del,float vn0,float eta0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEta(eta,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VnEtaDel2Cij(c11,c13,c33,vn,eta,del,vn0,eta0,del0,nnx*nnz);
 double val=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 delete []c11;delete []c13;delete []c33;
 return val; 
}

float objFunc21(float &datafit,float &reg,const float *d0,float *vn,float *eta,float *del,float vn0,float eta0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEta(eta,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VnEtaDel2Cij(c11,c13,c33,vn,eta,del,vn0,eta0,del0,nnx*nnz);
 datafit=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 float *tvn=new float[nnx*nnz](); ns_helixcon(tvn,vn,nnx*nnz,aa,lag,na,patch,false);
 float *teta=new float[nnx*nnz](); ns_helixcon(teta,eta,nnx*nnz,aa,lag,na,patch,false);
 float *tdel=new float[nnx*nnz](); ns_helixcon(tdel,del,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tvn,tvn,nnx*nnz)+dot_product(teta,teta,nnx*nnz)+dot_product(tdel,tdel,nnx*nnz)); 
 delete []c11;delete []c13;delete []c33;
 delete []tvn;delete []teta;delete []tdel;
 return datafit+reg; 
}

float objFunc3(const float *d0,float *vh,float *eps,float *eta,float vh0,float eps0,float eta0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEta(eta,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VhEpsEta2Cij(c11,c13,c33,vh,eps,eta,vh0,eps0,eta0,nnx*nnz);
 double val=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 delete []c11;delete []c13;delete []c33;
 return val; 
}

float objFunc31(float &datafit,float &reg,const float *d0,float *vh,float *eps,float *eta,float vh0,float eps0,float eta0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEta(eta,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VhEpsEta2Cij(c11,c13,c33,vh,eps,eta,vh0,eps0,eta0,nnx*nnz);
 datafit=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 float *tvh=new float[nnx*nnz](); ns_helixcon(tvh,vh,nnx*nnz,aa,lag,na,patch,false);
 float *teps=new float[nnx*nnz](); ns_helixcon(teps,eps,nnx*nnz,aa,lag,na,patch,false);
 float *teta=new float[nnx*nnz](); ns_helixcon(teta,eta,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tvh,tvh,nnx*nnz)+dot_product(teps,teps,nnx*nnz)+dot_product(teta,teta,nnx*nnz)); 
 delete []c11;delete []c13;delete []c33;
 delete []tvh;delete []teps;delete []teta;
 return datafit+reg; 
}

float objFunc4(const float *d0,float *v,float *vh,float *del,float v0,float vh0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVVhDel(v,vh,del,v0,vh0,del0,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VVhDel2Cij(c11,c13,c33,v,vh,del,v0,vh0,del0,nnx*nnz);
 double val=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 delete []c11;delete []c13;delete []c33;
 return val; 
}

float objFunc41(float &datafit,float &reg,const float *d0,float *v,float *vh,float *del,float v0,float vh0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVVhDel(v,vh,del,v0,vh0,del0,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VVhDel2Cij(c11,c13,c33,v,vh,del,v0,vh0,del0,nnx*nnz);
 datafit=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 float *tv=new float[nnx*nnz](); ns_helixcon(tv,v,nnx*nnz,aa,lag,na,patch,false);
 float *tvh=new float[nnx*nnz](); ns_helixcon(tvh,vh,nnx*nnz,aa,lag,na,patch,false);
 float *tdel=new float[nnx*nnz](); ns_helixcon(tdel,del,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tv,tv,nnx*nnz)+dot_product(tvh,tvh,nnx*nnz)+dot_product(tdel,tdel,nnx*nnz)); 
 delete []c11;delete []c13;delete []c33;
 delete []tv;delete []tvh;delete []tdel;
 return datafit+reg; 
}

float objFunc5(const float *d0,float *vn,float *vh,float *del,float vn0,float vh0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVnVh(vn,vh,vn0,vh0,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VnVhDel2Cij(c11,c13,c33,vn,vh,del,vn0,vh0,del0,nnx*nnz);
 double val=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 delete []c11;delete []c13;delete []c33;
 return val; 
}

float objFunc51(float &datafit,float &reg,const float *d0,float *vn,float *vh,float *del,float vn0,float vh0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVnVh(vn,vh,vn0,vh0,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VnVhDel2Cij(c11,c13,c33,vn,vh,del,vn0,vh0,del0,nnx*nnz);
 datafit=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 float *tvn=new float[nnx*nnz](); ns_helixcon(tvn,vn,nnx*nnz,aa,lag,na,patch,false);
 float *tvh=new float[nnx*nnz](); ns_helixcon(tvh,vh,nnx*nnz,aa,lag,na,patch,false);
 float *tdel=new float[nnx*nnz](); ns_helixcon(tdel,del,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tvn,tvn,nnx*nnz)+dot_product(tvh,tvh,nnx*nnz)+dot_product(tdel,tdel,nnx*nnz)); 
 delete []c11;delete []c13;delete []c33;
 delete []tvn;delete []tvh;delete []tdel;
 return datafit+reg; 
}

float objFunc6(const float *d0,float *v,float *vn,float *vh,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVnVh(vn,vh,1.,1.,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VVnVh2Cij(c11,c13,c33,v,vn,vh,nnx*nnz);
 double val=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 delete []c11;delete []c13;delete []c33;
 return val; 
}

float objFunc61(float &datafit,float &reg,const float *d0,float *v,float *vn,float *vh,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVnVh(vn,vh,1.,1.,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 VVnVh2Cij(c11,c13,c33,v,vn,vh,nnx*nnz);
 datafit=objFunc_f(d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 float *tv=new float[nnx*nnz](); ns_helixcon(tv,v,nnx*nnz,aa,lag,na,patch,false);
 float *tvn=new float[nnx*nnz](); ns_helixcon(tvn,vn,nnx*nnz,aa,lag,na,patch,false);
 float *tvh=new float[nnx*nnz](); ns_helixcon(tvh,vh,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tv,tv,nnx*nnz)+dot_product(tvn,tvn,nnx*nnz)+dot_product(tvh,tvh,nnx*nnz)); 
 delete []c11;delete []c13;delete []c33;
 delete []tvn;delete []tvh;delete []tv;
 return datafit+reg; 
}

float objFunc_f(const float *d0,const float *c11,const float *c13,const float *c33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 double val=0;
 fprintf(stderr,"Calculating obj func\n");

 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 int nnx=nx+2*npad,nnz=nz+2*npad;
 float dx2=dx*dx,dz2=dz*dz,dt2=dt*dt;
 
 int nGPUs;
 cudaGetDeviceCount(&nGPUs);
 //fprintf(stderr,"Total # GPUs = %d\n",nGPUs);

 float **d=new float*[nGPUs]();
 int **d_rloc=new int*[nGPUs]();
 float **d_c11=new float*[nGPUs]();
 float **d_c13=new float*[nGPUs]();
 float **d_c33=new float*[nGPUs]();
 float **d_taper=new float*[nGPUs]();
 float **d_d=new float*[nGPUs]();
 float **p0=new float*[nGPUs]();
 float **q0=new float*[nGPUs]();
 float **p1=new float*[nGPUs]();
 float **q1=new float*[nGPUs]();

 for(int i=0;i<nGPUs;++i){
  cudaSetDevice(i);
  
  init_coeffs();
 
  d[i]=new float[nr*nnt]();

  cudaMalloc(&d_rloc[i],2*nr*sizeof(int));
  cudaMemcpy(d_rloc[i],rloc,2*nr*sizeof(int),cudaMemcpyHostToDevice);
   
  cudaMalloc(&d_c11[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_c13[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_c33[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_c11[i],c11,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_c13[i],c13,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_c33[i],c33,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  
  cudaMalloc(&d_taper[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_taper[i],taper,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 
  cudaMalloc(&d_d[i],nr*sizeof(float));
 
  cudaMalloc(&p0[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&p1[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&q0[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&q1[i],nnx*nnz*sizeof(float)); 
 }
 
 int npasses=(ns+nGPUs-1)/nGPUs;
 int shotLeft=ns;

 for(int pass=0;pass<npasses;++pass){
  int nGPUsNeed=min(shotLeft,nGPUs);
  //fprintf(stderr,"Pass %d, # GPUs %d\n",pass,nGPUsNeed);
  #pragma omp parallel for num_threads(nGPUsNeed)
  for(int i=0;i<nGPUsNeed;++i){
   cudaSetDevice(i);

   int is=pass*nGPUs+i;
   int slocxz=sloc[0+is*2]+sloc[1+is*2]*nnx;

   dim3 block(BLOCK_DIM_X,BLOCK_DIM_Y);
   dim3 grid((nnx-2*RADIUS)/BLOCK_DIM_X,(nnz-2*RADIUS)/BLOCK_DIM_Y);

   cudaMemset(p0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(p1[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q1[i],0,nnx*nnz*sizeof(float));

   inject<<<1,1>>>(p1[i],q1[i],wavelet[0],slocxz,nnx,dt2);
  
   abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
   
   if(ratio==1){
    record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p1[i],q1[i],d_rloc[i],nr,nnx);
    cudaMemcpy(d[i]+nr,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
   }
   
   for(int it=2;it<nt;++it){
    //fprintf(stderr,"Time step it=%d\n",it);
  
    forwardCD<<<grid,block>>>(p0[i],q0[i],p1[i],q1[i],d_c11[i],d_c13[i],d_c33[i],dx2,dz2,dt2);
  
    inject<<<1,1>>>(p0[i],q0[i],wavelet[it-1],slocxz,nnx,dt2);
  
    abc<<<grid,block>>>(p0[i],q0[i],d_taper[i]);
    abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
    
    if(it%ratio==0){
     record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p0[i],q0[i],d_rloc[i],nr,nnx);
     cudaMemcpy(d[i]+(it/ratio)*nr,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
    }
  
    float *pt=p0[i]; 
    p0[i]=p1[i];
    p1[i]=pt;
    pt=q0[i];
    q0[i]=q1[i];
    q1[i]=pt;
   }
  }
  
  for(int i=0;i<nGPUsNeed;++i){
   int is=pass*nGPUs+i;
   #pragma omp parallel for num_threads(omp_get_num_threads()) reduction(+:val)
   for(size_t irt=0;irt<nr*nnt;++irt){
    val+=(d[i][irt]-d0[irt+is*nr*nnt])*(d[i][irt]-d0[irt+is*nr*nnt]);
   }
  }
  
  shotLeft-=nGPUsNeed;
 }

 for(int i=0;i<nGPUs;++i){
  cudaSetDevice(i);
  cudaFree(d_rloc[i]);
  cudaFree(d_c11[i]);cudaFree(d_c13[i]);cudaFree(d_c33[i]);
  cudaFree(d_taper[i]);
  cudaFree(d_d[i]);
  cudaFree(p0[i]);cudaFree(p1[i]);cudaFree(q0[i]);cudaFree(q1[i]);
 }
 
 delete []d_rloc;
 delete []d_c11;
 delete []d_c13;
 delete []d_c33;
 delete []d_taper;
 delete []d_d;
 delete []p0;
 delete []q0;
 delete []p1;
 delete []q1;
 
 return 0.5*val;
}
