#include <omp.h>
#include <cstdio>
#include "mylib.h"
#include "wave.h"
#include "check.h"
#include "conversions.h"
#include "kernels.h"
#include "helix.h"

float objFuncGradient0(float *gc11,float *gc13,float *gc33,const float *d0,float *c11,float *c13,float *c33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkCij(c11,c13,c33,nnx*nnz,m);
 float *tc11=new float[nnx*nnz](); scale(tc11,c11,1e6,nnx*nnz);
 float *tc13=new float[nnx*nnz](); scale(tc13,c13,1e6,nnx*nnz);
 float *tc33=new float[nnx*nnz](); scale(tc33,c33,1e6,nnx*nnz);
 double val=objFuncGradient_f(gc11,gc13,gc33,d0,tc11,tc13,tc33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 int n=npad+wbottom/dz+1;
 memset(gc11,0,n*nnx*sizeof(float)); scale(gc11,gc11,1e6,nnx*nnz);
 memset(gc13,0,n*nnx*sizeof(float)); scale(gc13,gc13,1e6,nnx*nnz);
 memset(gc33,0,n*nnx*sizeof(float)); scale(gc33,gc33,1e6,nnx*nnz);
 delete []tc11;delete []tc13;delete []tc33;
 return val; 
}

float objFuncGradient01(float &datafit,float &reg,float *gc11,float *gc13,float *gc33,const float *d0,float *c11,float *c13,float *c33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkCij(c11,c13,c33,nnx*nnz,m);
 float *tc11=new float[nnx*nnz](); scale(tc11,c11,1e6,nnx*nnz);
 float *tc13=new float[nnx*nnz](); scale(tc13,c13,1e6,nnx*nnz);
 float *tc33=new float[nnx*nnz](); scale(tc33,c33,1e6,nnx*nnz);
 double val=objFuncGradient_f(gc11,gc13,gc33,d0,tc11,tc13,tc33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 scale(gc11,gc11,1e6,nnx*nnz);
 scale(gc13,gc13,1e6,nnx*nnz);
 scale(gc33,gc33,1e6,nnx*nnz);
 delete []tc11;delete []tc13;delete []tc33;
 float *t11=new float[nnx*nnz](); ns_helixcon(t11,c11,nnx*nnz,aa,lag,na,patch,false);
 float *t13=new float[nnx*nnz](); ns_helixcon(t13,c13,nnx*nnz,aa,lag,na,patch,false);
 float *t33=new float[nnx*nnz](); ns_helixcon(t33,c33,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(t11,t11,nnx*nnz)+dot_product(t13,t13,nnx*nnz)+dot_product(t33,t33,nnx*nnz)); 
 float *tt11=new float[nnx*nnz](); ns_helixcon(t11,tt11,nnx*nnz,aa,lag,na,patch,true);
 float *tt13=new float[nnx*nnz](); ns_helixcon(t13,tt13,nnx*nnz,aa,lag,na,patch,true);
 float *tt33=new float[nnx*nnz](); ns_helixcon(t33,tt33,nnx*nnz,aa,lag,na,patch,true);
 scale_add(gc11,tt11,alpha,nnx*nnz);
 scale_add(gc13,tt13,alpha,nnx*nnz);
 scale_add(gc33,tt33,alpha,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gc11,0,n*nnx*sizeof(float)); 
 memset(gc13,0,n*nnx*sizeof(float)); 
 memset(gc33,0,n*nnx*sizeof(float)); 
 delete []t11;delete []t13;delete []t33;
 delete []tt11;delete []tt13;delete []tt33;
 return datafit+reg; 
}

float objFuncGradient1(float *gv,float *geps,float *gdel,const float *d0,float *v,float *eps,float *del,float v0,float eps0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEpsDel(eps,del,eps0,del0,nnx*nnz,m);
// for(int iz=0;iz<nnz;++iz){
//  for(int ix=0;ix<nnx;++ix){
//   if(m[ix+iz*nnx]==1){
//    del[ix+iz*nnx]=(del[ix-1+(iz-1)*nnx]+del[ix+(iz-1)*nnx]+del[ix+1+(iz-1)*nnx]+
//                    del[ix-1+iz*nnx]+del[ix+iz*nnx]+del[ix+1+iz*nnx]+
//	            	del[ix-1+(iz+1)*nnx]+del[ix+(iz+1)*nnx]+del[ix+1+(iz+1)*nnx])/9.;
//    eps[ix+iz*nnx]=(eps[ix-1+(iz-1)*nnx]+eps[ix+(iz-1)*nnx]+eps[ix+1+(iz-1)*nnx]+
//                    eps[ix-1+iz*nnx]+eps[ix+iz*nnx]+eps[ix+1+iz*nnx]+
//	            	eps[ix-1+(iz+1)*nnx]+eps[ix+(iz+1)*nnx]+eps[ix+1+(iz+1)*nnx])/9.;
//   }					
//  }
// }
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VEpsDel2Cij(c11,c13,c33,v,eps,del,v0,eps0,del0,nnx*nnz);
 double val=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVEpsDel(gv,geps,gdel,gc11,gc13,gc33,v,eps,del,v0,eps0,del0,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gv,0,n*nnx*sizeof(float));
 memset(geps,0,n*nnx*sizeof(float));
 memset(gdel,0,n*nnx*sizeof(float));
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 return val; 
}

float objFuncGradient11(float &datafit,float &reg,float *gv,float *geps,float *gdel,const float *d0,float *v,float *eps,float *del,float v0,float eps0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEpsDel(eps,del,eps0,del0,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VEpsDel2Cij(c11,c13,c33,v,eps,del,v0,eps0,del0,nnx*nnz);
 datafit=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVEpsDel(gv,geps,gdel,gc11,gc13,gc33,v,eps,del,v0,eps0,del0,nnx*nnz);
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 float *tv=new float[nnx*nnz](); ns_helixcon(tv,v,nnx*nnz,aa,lag,na,patch,false);
 float *teps=new float[nnx*nnz](); ns_helixcon(teps,eps,nnx*nnz,aa,lag,na,patch,false);
 float *tdel=new float[nnx*nnz](); ns_helixcon(tdel,del,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tv,tv,nnx*nnz)+dot_product(teps,teps,nnx*nnz)+dot_product(tdel,tdel,nnx*nnz)); 
 float *ttv=new float[nnx*nnz](); ns_helixcon(tv,ttv,nnx*nnz,aa,lag,na,patch,true);
 float *tteps=new float[nnx*nnz](); ns_helixcon(teps,tteps,nnx*nnz,aa,lag,na,patch,true);
 float *ttdel=new float[nnx*nnz](); ns_helixcon(tdel,ttdel,nnx*nnz,aa,lag,na,patch,true);
 scale_add(gv,ttv,alpha,nnx*nnz);
 scale_add(geps,tteps,alpha,nnx*nnz);
 scale_add(gdel,ttdel,alpha,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gv,0,n*nnx*sizeof(float));
 memset(geps,0,n*nnx*sizeof(float));
 memset(gdel,0,n*nnx*sizeof(float));
 delete []tv;delete []teps;delete []tdel;
 delete []ttv;delete []tteps;delete []ttdel;
 return datafit+reg; 
}

float objFuncGradient2(float *gvn,float *geta,float *gdel,const float *d0,float *vn,float *eta,float *del,float vn0,float eta0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEta(eta,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VnEtaDel2Cij(c11,c13,c33,vn,eta,del,vn0,eta0,del0,nnx*nnz);
 double val=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVnEtaDel(gvn,geta,gdel,gc11,gc13,gc33,vn,eta,del,vn0,eta0,del0,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gvn,0,n*nnx*sizeof(float));
 memset(geta,0,n*nnx*sizeof(float));
 memset(gdel,0,n*nnx*sizeof(float));
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 return val; 
}

float objFuncGradient21(float &datafit,float &reg,float *gvn,float *geta,float *gdel,const float *d0,float *vn,float *eta,float *del,float vn0,float eta0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEta(eta,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VnEtaDel2Cij(c11,c13,c33,vn,eta,del,vn0,eta0,del0,nnx*nnz);
 datafit=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVnEtaDel(gvn,geta,gdel,gc11,gc13,gc33,vn,eta,del,vn0,eta0,del0,nnx*nnz);
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 float *tvn=new float[nnx*nnz](); ns_helixcon(tvn,vn,nnx*nnz,aa,lag,na,patch,false);
 float *teta=new float[nnx*nnz](); ns_helixcon(teta,eta,nnx*nnz,aa,lag,na,patch,false);
 float *tdel=new float[nnx*nnz](); ns_helixcon(tdel,del,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tvn,tvn,nnx*nnz)+dot_product(teta,teta,nnx*nnz)+dot_product(tdel,tdel,nnx*nnz)); 
 float *ttvn=new float[nnx*nnz](); ns_helixcon(tvn,ttvn,nnx*nnz,aa,lag,na,patch,true);
 float *tteta=new float[nnx*nnz](); ns_helixcon(teta,tteta,nnx*nnz,aa,lag,na,patch,true);
 float *ttdel=new float[nnx*nnz](); ns_helixcon(tdel,ttdel,nnx*nnz,aa,lag,na,patch,true);
 scale_add(gvn,ttvn,alpha,nnx*nnz);
 scale_add(geta,tteta,alpha,nnx*nnz);
 scale_add(gdel,ttdel,alpha,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gvn,0,n*nnx*sizeof(float));
 memset(geta,0,n*nnx*sizeof(float));
 memset(gdel,0,n*nnx*sizeof(float));
 delete []tvn;delete []teta;delete []tdel;
 delete []ttvn;delete []tteta;delete []ttdel;
 return datafit+reg; 
}

float objFuncGradient3(float *gvh,float *geps,float *geta,const float *d0,float *vh,float *eps,float *eta,float vh0,float eps0,float eta0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEta(eta,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VhEpsEta2Cij(c11,c13,c33,vh,eps,eta,vh0,eps0,eta0,nnx*nnz);
 double val=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVhEpsEta(gvh,geps,geta,gc11,gc13,gc33,vh,eps,eta,vh0,eps0,eta0,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gvh,0,n*nnx*sizeof(float));
 memset(geps,0,n*nnx*sizeof(float));
 memset(geta,0,n*nnx*sizeof(float));
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 return val; 
}

float objFuncGradient31(float &datafit,float &reg,float *gvh,float *geps,float *geta,const float *d0,float *vh,float *eps,float *eta,float vh0,float eps0,float eta0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkEta(eta,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VhEpsEta2Cij(c11,c13,c33,vh,eps,eta,vh0,eps0,eta0,nnx*nnz);
 datafit=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVhEpsEta(gvh,geps,geta,gc11,gc13,gc33,vh,eps,eta,vh0,eps0,eta0,nnx*nnz);
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 float *tvh=new float[nnx*nnz](); ns_helixcon(tvh,vh,nnx*nnz,aa,lag,na,patch,false);
 float *teps=new float[nnx*nnz](); ns_helixcon(teps,eps,nnx*nnz,aa,lag,na,patch,false);
 float *teta=new float[nnx*nnz](); ns_helixcon(teta,eta,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tvh,tvh,nnx*nnz)+dot_product(teps,teps,nnx*nnz)+dot_product(teta,teta,nnx*nnz)); 
 float *ttvh=new float[nnx*nnz](); ns_helixcon(tvh,ttvh,nnx*nnz,aa,lag,na,patch,true);
 float *tteps=new float[nnx*nnz](); ns_helixcon(teps,tteps,nnx*nnz,aa,lag,na,patch,true);
 float *tteta=new float[nnx*nnz](); ns_helixcon(teta,tteta,nnx*nnz,aa,lag,na,patch,true);
 scale_add(gvh,ttvh,alpha,nnx*nnz);
 scale_add(geps,tteps,alpha,nnx*nnz);
 scale_add(geta,tteta,alpha,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gvh,0,n*nnx*sizeof(float));
 memset(geps,0,n*nnx*sizeof(float));
 memset(geta,0,n*nnx*sizeof(float));
 delete []tvh;delete []teps;delete []teta;
 delete []ttvh;delete []tteps;delete []tteta;
 return datafit+reg; 
}

float objFuncGradient4(float *gv,float *gvh,float *gdel,const float *d0,float *v,float *vh,float *del,float v0,float vh0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVVhDel(v,vh,del,v0,vh0,del0,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VVhDel2Cij(c11,c13,c33,v,vh,del,v0,vh0,del0,nnx*nnz);
 double val=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVVhDel(gv,gvh,gdel,gc11,gc13,gc33,v,vh,del,v0,vh0,del0,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gv,0,n*nnx*sizeof(float));
 memset(gvh,0,n*nnx*sizeof(float));
 memset(gdel,0,n*nnx*sizeof(float));
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 return val; 
}

float objFuncGradient41(float &datafit,float &reg,float *gv,float *gvh,float *gdel,const float *d0,float *v,float *vh,float *del,float v0,float vh0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVVhDel(v,vh,del,v0,vh0,del0,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VVhDel2Cij(c11,c13,c33,v,vh,del,v0,vh0,del0,nnx*nnz);
 datafit=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVVhDel(gv,gvh,gdel,gc11,gc13,gc33,v,vh,del,v0,vh0,del0,nnx*nnz);
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 float *tv=new float[nnx*nnz](); ns_helixcon(tv,v,nnx*nnz,aa,lag,na,patch,false);
 float *tvh=new float[nnx*nnz](); ns_helixcon(tvh,vh,nnx*nnz,aa,lag,na,patch,false);
 float *tdel=new float[nnx*nnz](); ns_helixcon(tdel,del,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tv,tv,nnx*nnz)+dot_product(tvh,tvh,nnx*nnz)+dot_product(tdel,tdel,nnx*nnz)); 
 float *ttv=new float[nnx*nnz](); ns_helixcon(tv,ttv,nnx*nnz,aa,lag,na,patch,true);
 float *ttvh=new float[nnx*nnz](); ns_helixcon(tvh,ttvh,nnx*nnz,aa,lag,na,patch,true);
 float *ttdel=new float[nnx*nnz](); ns_helixcon(tdel,ttdel,nnx*nnz,aa,lag,na,patch,true);
 scale_add(gv,ttv,alpha,nnx*nnz);
 scale_add(gvh,ttvh,alpha,nnx*nnz);
 scale_add(gdel,ttdel,alpha,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gv,0,n*nnx*sizeof(float));
 memset(gvh,0,n*nnx*sizeof(float));
 memset(gdel,0,n*nnx*sizeof(float));
 delete []tv;delete []tvh;delete []tdel;
 delete []ttv;delete []ttvh;delete []ttdel;
 return datafit+reg; 
}

float objFuncGradient5(float *gvn,float *gvh,float *gdel,const float *d0,float *vn,float *vh,float *del,float vn0,float vh0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVnVh(vn,vh,vn0,vh0,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VnVhDel2Cij(c11,c13,c33,vn,vh,del,vn0,vh0,del0,nnx*nnz);
 double val=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVnVhDel(gvn,gvh,gdel,gc11,gc13,gc33,vn,vh,del,vn0,vh0,del0,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gvn,0,n*nnx*sizeof(float));
 memset(gvh,0,n*nnx*sizeof(float));
 memset(gdel,0,n*nnx*sizeof(float));
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 return val; 
}

float objFuncGradient51(float &datafit,float &reg,float *gvn,float *gvh,float *gdel,const float *d0,float *vn,float *vh,float *del,float vn0,float vh0,float del0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVnVh(vn,vh,vn0,vh0,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VnVhDel2Cij(c11,c13,c33,vn,vh,del,vn0,vh0,del0,nnx*nnz);
 datafit=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVnVhDel(gvn,gvh,gdel,gc11,gc13,gc33,vn,vh,del,vn0,vh0,del0,nnx*nnz);
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 float *tvn=new float[nnx*nnz](); ns_helixcon(tvn,vn,nnx*nnz,aa,lag,na,patch,false);
 float *tvh=new float[nnx*nnz](); ns_helixcon(tvh,vh,nnx*nnz,aa,lag,na,patch,false);
 float *tdel=new float[nnx*nnz](); ns_helixcon(tdel,del,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tvn,tvn,nnx*nnz)+dot_product(tvh,tvh,nnx*nnz)+dot_product(tdel,tdel,nnx*nnz)); 
 float *ttvn=new float[nnx*nnz](); ns_helixcon(tvn,ttvn,nnx*nnz,aa,lag,na,patch,true);
 float *ttvh=new float[nnx*nnz](); ns_helixcon(tvh,ttvh,nnx*nnz,aa,lag,na,patch,true);
 float *ttdel=new float[nnx*nnz](); ns_helixcon(tdel,ttdel,nnx*nnz,aa,lag,na,patch,true);
 scale_add(gvn,ttvn,alpha,nnx*nnz);
 scale_add(gvh,ttvh,alpha,nnx*nnz);
 scale_add(gdel,ttdel,alpha,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gvn,0,n*nnx*sizeof(float));
 memset(gvh,0,n*nnx*sizeof(float));
 memset(gdel,0,n*nnx*sizeof(float));
 delete []tvn;delete []tvh;delete []tdel;
 delete []ttvn;delete []ttvh;delete []ttdel;
 return datafit+reg; 
}

float objFuncGradient6(float *gv,float *gvn,float *gvh,const float *d0,float *v,float *vn,float *vh,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVnVh(vn,vh,1.,1.,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VVnVh2Cij(c11,c13,c33,v,vn,vh,nnx*nnz);
 double val=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVVnVh(gv,gvn,gvh,gc11,gc13,gc33,v,vn,vh,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gv,0,n*nnx*sizeof(float));
 memset(gvn,0,n*nnx*sizeof(float));
 memset(gvh,0,n*nnx*sizeof(float));
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 return val; 
}

float objFuncGradient61(float &datafit,float &reg,float *gv,float *gvn,float *gvh,const float *d0,float *v,float *vn,float *vh,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate,float wbottom,float *m,float alpha,float **aa,int **lag,int *na,int *patch){
 int nnx=nx+2*npad,nnz=nz+2*npad;
 checkVnVh(vn,vh,1.,1.,nnx*nnz,m);
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 VVnVh2Cij(c11,c13,c33,v,vn,vh,nnx*nnz);
 datafit=objFuncGradient_f(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 GradCij2GradVVnVh(gv,gvn,gvh,gc11,gc13,gc33,v,vn,vh,nnx*nnz);
 delete []c11;delete []c13;delete []c33;
 delete []gc11;delete []gc13;delete []gc33;
 float *tv=new float[nnx*nnz](); ns_helixcon(tv,v,nnx*nnz,aa,lag,na,patch,false);
 float *tvn=new float[nnx*nnz](); ns_helixcon(tvn,vn,nnx*nnz,aa,lag,na,patch,false);
 float *tvh=new float[nnx*nnz](); ns_helixcon(tvh,vh,nnx*nnz,aa,lag,na,patch,false);
 reg=0.5*alpha*(dot_product(tv,tv,nnx*nnz)+dot_product(tvn,tvn,nnx*nnz)+dot_product(tvh,tvh,nnx*nnz)); 
 float *ttv=new float[nnx*nnz](); ns_helixcon(tv,ttv,nnx*nnz,aa,lag,na,patch,true);
 float *ttvn=new float[nnx*nnz](); ns_helixcon(tvn,ttvn,nnx*nnz,aa,lag,na,patch,true);
 float *ttvh=new float[nnx*nnz](); ns_helixcon(tvh,ttvh,nnx*nnz,aa,lag,na,patch,true);
 scale_add(gv,ttv,alpha,nnx*nnz);
 scale_add(gvn,ttvn,alpha,nnx*nnz);
 scale_add(gvh,ttvh,alpha,nnx*nnz);
 int n=npad+wbottom/dz+1;
 memset(gv,0,n*nnx*sizeof(float));
 memset(gvn,0,n*nnx*sizeof(float));
 memset(gvh,0,n*nnx*sizeof(float));
 delete []tv;delete []tvn;delete []tvh;
 delete []ttv;delete []ttvn;delete []ttvh;
 return datafit+reg; 
}

float objFuncGradient_f(float *gc11,float *gc13,float *gc33,const float *d0,const float *c11,const float *c13,const float *c33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 double val=0;
 fprintf(stderr,"Calculate obj func and grad\n");
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 int nnx=nx+2*npad,nnz=nz+2*npad;
 float dx2=dx*dx,dz2=dz*dz,dt2=dt*dt;
 
 memset(gc11,0,nnx*nnz*sizeof(float));
 memset(gc13,0,nnx*nnz*sizeof(float));
 memset(gc33,0,nnx*nnz*sizeof(float));
 
 int nGPUs;
 cudaGetDeviceCount(&nGPUs);
 //fprintf(stderr,"Total # GPUs = %d\n",nGPUs);

 float **d=new float*[nGPUs]();
 int **d_rloc=new int*[nGPUs]();
 float **d_c11=new float*[nGPUs]();
 float **d_c13=new float*[nGPUs]();
 float **d_c33=new float*[nGPUs]();
 float **d_taper=new float*[nGPUs]();
 float **d_d=new float*[nGPUs]();
 float **p0=new float*[nGPUs]();
 float **q0=new float*[nGPUs]();
 float **p1=new float*[nGPUs]();
 float **q1=new float*[nGPUs]();
 float **tp=new float*[nGPUs]();
 float **tq=new float*[nGPUs]();
 float **Dp=new float*[nGPUs]();
 float **Dq=new float*[nGPUs]();
 float **d_gc11=new float*[nGPUs]();
 float **d_gc13=new float*[nGPUs]();
 float **d_gc33=new float*[nGPUs]();
 float **d_d0=new float*[nGPUs]();
 float **d_d1=new float*[nGPUs]();
 float **d_d00=new float*[nGPUs]();
 float **d_d01=new float*[nGPUs]();
 float **d_Dp0=new float*[nGPUs]();
 float **d_Dp1=new float*[nGPUs]();
 float **d_Dq0=new float*[nGPUs]();
 float **d_Dq1=new float*[nGPUs]();
 float **tgc11=new float*[nGPUs]();
 float **tgc13=new float*[nGPUs]();
 float **tgc33=new float*[nGPUs]();
 
 for(int i=0;i<nGPUs;++i){
  cudaSetDevice(i);
  
  init_coeffs();
 
  d[i]=new float[nr*nnt]();
  
  cudaMalloc(&d_rloc[i],2*nr*sizeof(int));
  cudaMemcpy(d_rloc[i],rloc,2*nr*sizeof(int),cudaMemcpyHostToDevice);
   
  cudaMalloc(&d_c11[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_c13[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_c33[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_c11[i],c11,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_c13[i],c13,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_c33[i],c33,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  
  cudaMalloc(&d_taper[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_taper[i],taper,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 
  cudaMalloc(&d_d[i],nr*sizeof(float));
 
  cudaMalloc(&p0[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&p1[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&q0[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&q1[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&tp[i],nnx*nnz*sizeof(float)); cudaMemset(tp[i],0,nnx*nnz*sizeof(float));
  cudaMalloc(&tq[i],nnx*nnz*sizeof(float)); cudaMemset(tq[i],0,nnx*nnz*sizeof(float));
  
  Dp[i]=new float[nnx*nnz*nnt]();
  Dq[i]=new float[nnx*nnz*nnt]();
  
  cudaMalloc(&d_gc11[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_gc13[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_gc33[i],nnx*nnz*sizeof(float));
  
  cudaMalloc(&d_d0[i],nr*sizeof(float));
  cudaMalloc(&d_d1[i],nr*sizeof(float));
  cudaMalloc(&d_d00[i],nr*sizeof(float));
  cudaMalloc(&d_d01[i],nr*sizeof(float));
 
  cudaMalloc(&d_Dp0[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_Dq0[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_Dp1[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_Dq1[i],nnx*nnz*sizeof(float));
  
  tgc11[i]=new float[nnx*nnz]();
  tgc13[i]=new float[nnx*nnz]();
  tgc33[i]=new float[nnx*nnz]();
 }

 int npasses=(ns+nGPUs-1)/nGPUs;
 int shotLeft=ns;

 for(int pass=0;pass<npasses;++pass){
  int nGPUsNeed=min(shotLeft,nGPUs);
  //fprintf(stderr,"Pass %d, # GPUs %d\n",pass,nGPUsNeed);
  
  #pragma omp parallel for num_threads(nGPUsNeed)
  for(int i=0;i<nGPUsNeed;++i){
   cudaSetDevice(i);

   int is=pass*nGPUs+i;
   int slocxz=sloc[0+is*2]+sloc[1+is*2]*nnx;

   dim3 block(BLOCK_DIM_X,BLOCK_DIM_Y);
   dim3 grid((nnx-2*RADIUS)/BLOCK_DIM_X,(nnz-2*RADIUS)/BLOCK_DIM_Y);

   cudaMemset(p0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(p1[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q1[i],0,nnx*nnz*sizeof(float));
 
   inject<<<1,1>>>(p1[i],q1[i],wavelet[0],slocxz,nnx,dt2);
  
   abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
  
   D<<<grid,block>>>(tp[i],tq[i],p1[i],q1[i],dx2,dz2);
  
   if(ratio==1){
    record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p1[i],q1[i],d_rloc[i],nr,nnx);
    cudaMemcpy(d[i]+nr,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
    cudaMemcpy(Dp[i]+nnx*nnz,tp[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
    cudaMemcpy(Dq[i]+nnx*nnz,tq[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
   }
   
   for(int it=2;it<nt;++it){
    //fprintf(stderr,"Time step it=%d\n",it);
  
    forwardC<<<grid,block>>>(p0[i],q0[i],p1[i],q1[i],tp[i],tq[i],d_c11[i],d_c13[i],d_c33[i],dt2);
  
    inject<<<1,1>>>(p0[i],q0[i],wavelet[it-1],slocxz,nnx,dt2);
  
    abc<<<grid,block>>>(p0[i],q0[i],d_taper[i]);
    abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
    
    D<<<grid,block>>>(tp[i],tq[i],p0[i],q0[i],dx2,dz2);
  
    if(it%ratio==0){
     record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p0[i],q0[i],d_rloc[i],nr,nnx);
     cudaMemcpy(d[i]+(it/ratio)*nr,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
     cudaMemcpy(Dp[i]+(it/ratio)*nnx*nnz,tp[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
     cudaMemcpy(Dq[i]+(it/ratio)*nnx*nnz,tq[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
    }
  
    float *pt=p0[i]; 
    p0[i]=p1[i];
    p1[i]=pt;
    pt=q0[i];
    q0[i]=q1[i];
    q1[i]=pt;
   }
 
   cudaMemset(p0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(p1[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q1[i],0,nnx*nnz*sizeof(float));
  
   cudaMemset(d_gc11[i],0,nnx*nnz*sizeof(float));
   cudaMemset(d_gc13[i],0,nnx*nnz*sizeof(float));
   cudaMemset(d_gc33[i],0,nnx*nnz*sizeof(float));
   
   cudaMemcpy(d_d00[i],d0+(nt-1)/ratio*nr+is*nr*nnt,nr*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_d0[i],d[i]+(nt-1)/ratio*nr,nr*sizeof(float),cudaMemcpyHostToDevice);
   inject<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(p0[i],q0[i],d_d00[i],d_d01[i],d_d0[i],d_d1[i],0.,d_rloc[i],nr,nnx,dt2);
  
   abc<<<grid,block>>>(p0[i],q0[i],d_taper[i]);
  
   float f=(nt-2.)/ratio;
   int i1=f,i2=i1+1;
   cudaMemcpy(d_Dp0[i],Dp[i]+i1*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_Dq0[i],Dq[i]+i1*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_Dp1[i],Dp[i]+i2*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_Dq1[i],Dq[i]+i2*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   f=f-i1;
   imaging<<<grid,block>>>(d_gc11[i],d_gc13[i],d_gc33[i],p0[i],q0[i],d_Dp0[i],d_Dq0[i],d_Dp1[i],d_Dq1[i],f);
  
   for(int it=nt-3;it>=0;--it){
    //fprintf(stderr,"Time step it=%d\n",it);
    
    C<<<grid,block>>>(tp[i],tq[i],p0[i],q0[i],d_c11[i],d_c13[i],d_c33[i]);
  
    backwardD<<<grid,block>>>(p1[i],q1[i],p0[i],q0[i],tp[i],tq[i],dx2,dz2,dt2);
  
    f=(it+1.)/ratio;
    i1=f;
    if((it+2)%ratio==0){
     cudaMemcpy(d_d1[i],d[i]+i1*nr,nr*sizeof(float),cudaMemcpyHostToDevice);
     cudaMemcpy(d_d01[i],d0+i1*nr+is*nr*nnt,nr*sizeof(float),cudaMemcpyHostToDevice);
     float *pt=d_d0[i]; 
     d_d0[i]=d_d1[i];
     d_d1[i]=pt;
     pt=d_d00[i];
     d_d00[i]=d_d01[i];
     d_d01[i]=pt;
    }
    f=f-i1;
    inject<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(p1[i],q1[i],d_d00[i],d_d01[i],d_d0[i],d_d1[i],f,d_rloc[i],nr,nnx,dt2);
  
    abc<<<grid,block>>>(p0[i],q0[i],d_taper[i]);
    abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
  
    f=(float)it/ratio;
    i1=f;
    if((it+1)%ratio==0){
     cudaMemcpy(d_Dp1[i],Dp[i]+i1*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
     cudaMemcpy(d_Dq1[i],Dq[i]+i1*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
     float *pt=d_Dp0[i]; 
     d_Dp0[i]=d_Dp1[i];
     d_Dp1[i]=pt;
     pt=d_Dq0[i];
     d_Dq0[i]=d_Dq1[i];
     d_Dq1[i]=pt;
    }
    f=f-i1;
    imaging<<<grid,block>>>(d_gc11[i],d_gc13[i],d_gc33[i],p1[i],q1[i],d_Dp0[i],d_Dq0[i],d_Dp1[i],d_Dq1[i],f);
  
    float *pt=p0[i]; 
    p0[i]=p1[i];
    p1[i]=pt;
    pt=q0[i];
    q0[i]=q1[i];
    q1[i]=pt;
   }
   cudaMemcpy(tgc11[i],d_gc11[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
   cudaMemcpy(tgc13[i],d_gc13[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
   cudaMemcpy(tgc33[i],d_gc33[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
  }
  
  for(int i=0;i<nGPUsNeed;++i){
   int is=pass*nGPUs+i;
   #pragma omp parallel for num_threads(omp_get_num_threads()) reduction(+:val)
   for(int irt=0;irt<nr*nnt;++irt){
    val+=(d[i][irt]-d0[irt+is*nr*nnt])*(d[i][irt]-d0[irt+is*nr*nnt]);
   }
   
   #pragma omp parallel for num_threads(omp_get_num_threads()) shared(i)
   for(int ixz=0;ixz<nnx*nnz;++ixz){
    gc11[ixz]+=tgc11[i][ixz];
    gc13[ixz]+=tgc13[i][ixz];
    gc33[ixz]+=tgc33[i][ixz];
   }
  }
  
  shotLeft-=nGPUsNeed;
 }

 for(int i=0;i<nGPUs;++i){
  cudaSetDevice(i);
  delete []d[i];
  cudaFree(d_rloc[i]);
  cudaFree(d_c11[i]);cudaFree(d_c13[i]);cudaFree(d_c33[i]); 
  cudaFree(d_taper[i]);
  cudaFree(d_d[i]);
  cudaFree(p0[i]);cudaFree(p1[i]);cudaFree(q0[i]);cudaFree(q1[i]);cudaFree(tp[i]);cudaFree(tq[i]);
  delete []Dp[i];delete []Dq[i]; 
  cudaFree(d_gc11[i]);cudaFree(d_gc13[i]);cudaFree(d_gc33[i]);
  cudaFree(d_d0[i]);cudaFree(d_d1[i]);cudaFree(d_d00[i]);cudaFree(d_d01[i]);
  cudaFree(d_Dp0[i]);cudaFree(d_Dp1[i]);cudaFree(d_Dq0[i]);cudaFree(d_Dq1[i]);
  delete []tgc11[i];delete []tgc13[i];delete []tgc33[i];
 }
 
 delete []d;
 delete []d_rloc;
 delete []d_c11;delete []d_c13;delete []d_c33;
 delete []d_taper;
 delete []d_d;
 delete []p0;delete []p1;delete []q0;delete []q1;delete []tp;delete []tq;
 delete []Dp;delete []Dq;
 delete []d_gc11;delete []d_gc13;delete []d_gc33;
 delete []d_d0;delete []d_d1;delete []d_d00;delete []d_d01;
 delete []d_Dp0;delete []d_Dp1;delete []d_Dq0;delete []d_Dq1;
 delete []tgc11;delete []tgc13;delete []tgc33;

 return 0.5*val;
}
