#include <cmath>
#include <omp.h>
#include <cstdio>
#include "myio.h"
#include "mylib.h"
#include "init.h"
#include "wave.h"

using namespace std;

int main(int argc,char **argv){
 myio_init(argc,argv);

int nnx,nnz,nx,nz,nt,npad;
 float ox,oz,ot,dx,dz,dt;
 init2d(nnx,nnz,nx,nz,nt,dx,dz,dt,ox,oz,ot,npad);

 float f;
 get_param("f",f);
 float *wavelet=new float[nt]();
 ricker(wavelet,f,nt,dt);

 float *v=new float[nnx*nnz]();
 float *eps=new float[nnx*nnz]();
 float *del=new float[nnx*nnz]();
 
 init_model("v",v,nx,nz,npad); 
 init_model("eps",eps,nx,nz,npad); 
 init_model("del",del,nx,nz,npad); 

 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();

 #pragma omp parallel for default(shared) 
 for(size_t i=0;i<nnx*nnz;++i){
  c11[i]=v[i]*v[i]*(1.+2.*eps[i]);
  c13[i]=v[i]*v[i]*sqrt(1.+2.*del[i]);
  c33[i]=v[i]*v[i];
 }
 
 int ns;
 float os,ds,zs;
 init_shot(ns,ds,os,zs);
 int *sloc=new int[2*ns]();
 init_shot_loc(sloc,ns,ds,os,zs,dx,dz,ox,oz,npad);
 
 int nr;
 float o_r,dr,zr;
 init_rec(nr,dr,o_r,zr);
 int *rloc=new int[2*nr]();
 init_rec_loc(rloc,nr,dr,o_r,zr,dx,dz,ox,oz,npad);
 
 float rate;
 get_param("rate",rate);
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 float *d0=new float[nr*nnt*ns]();
 read("data",d0,nr*nnt*ns);
 
 float *taper=new float[nnx*nnz]();
 init_abc(taper,nx,nz,npad);
 
 float *image=new float[nnx*nnz]();
 float *image1=new float[nnx*nnz]();
 
 rtm_f(image,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 
 for(int iz=1;iz<nnz-1;++iz){
  for(int ix=1;ix<nnx-1;++ix){
   image1[ix+iz*nnx]=image[ix-1+iz*nnx]+image[ix+1+iz*nnx]+image[ix+(iz-1)*nnx]+image[ix+(iz+1)*nnx]-4.*image[ix+iz*nnx];
  }
 }

 float wbottom; get_param("wbottom",wbottom);
 int n=npad+wbottom/dz+1;
 memset(image,0,n*nnx*sizeof(float)); 

 write("image",image1,nnx*nnz);
 to_header("image","n1",nnx,"o1",-dx*npad,"d1",dx);
 to_header("image","n2",nnz,"o2",-dz*npad,"d2",dz);
 
 delete []wavelet;delete []v;delete []eps;delete []del;delete []d0;delete []sloc;delete []rloc;delete []taper;
 delete []c11;delete []c13;delete []c33;delete []image;delete []image1;

 myio_close();
 return 0;
}
